
import numpy as np, itertools, collections, dataclasses

from McUtils.Numputils import SparseArray
import McUtils.Numputils as nput
from McUtils.Data import UnitsData
from McUtils.Scaffolding import Logger, NullLogger, Checkpointer

from ..Spectra import DiscreteSpectrum
from ..BasisReps import (
    BasisStateSpace, BasisMultiStateSpace, SelectionRuleStateSpace, HarmonicOscillatorProductBasis,
    StateMaker
)
from .Common import PerturbationTheoryException, _safe_dot

__all__ = [
    "PerturbationTheoryCorrections",
    "AnalyticPerturbationTheoryCorrections",
    "BasicAPTCorrections"
]


class PerturbationTheoryCorrections:
    """
    Represents a set of corrections from perturbation theory.
    Can be used to correct other operators in the basis of the original calculation.

    """
    def __init__(self,
                 states,
                 coupled_states,
                 total_basis,
                 energy_corrs,
                 wfn_corrections,
                 all_energy_corrections=None,
                 degenerate_states=None,
                 degenerate_transformation=None,
                 degenerate_energies=None,
                 degenerate_hamiltonians=None,
                 nondeg_hamiltonian_precision=3,
                 logger=None
                 ):
        """
        :param states:
        :type states: BasisStateSpace
        :param coupled_states:
        :type coupled_states: BasisMultiStateSpace
        :param total_basis:
        :type total_basis: BasisMultiStateSpace
        :param energy_corrs:
        :type energy_corrs: np.ndarray
        :param wfn_corrections:
        :type wfn_corrections: Iterable[SparseArray]
        :param degenerate_states:
        :type degenerate_states: None | np.ndarray
        :param degenerate_transformation:
        :type degenerate_transformation: None | np.ndarray
        :param degenerate_energies:
        :type degenerate_energies: None | np.ndarray
        """
        self.states = states
        self.coupled_states = coupled_states
        self.total_basis = total_basis
        self.energy_corrs = energy_corrs
        self.all_energy_corrs = all_energy_corrections
        self.wfn_corrections = wfn_corrections
        self.degenerate_states = degenerate_states
        self.degenerate_transf = degenerate_transformation
        self.degenerate_energies = degenerate_energies
        self.degenerate_hamiltonians = degenerate_hamiltonians
        self.logger = logger
        self.nondeg_hamiltonian_precision = nondeg_hamiltonian_precision

    @classmethod
    def from_dicts(cls,
                   states,
                   corrections,
                   **opts
                   ):
        """
        :param states: a dict with the states described by the corrections, the set of states coupled, and the size of the overall basis
        :type states: dict
        :param corrections: the corrections generated, including the corrections for the energies, wavefunctions, and a transformation from degenerate PT
        :type corrections: dict
        """
        state_space = states['states']
        coupled_states = states['coupled_states']
        total_basis = states['total_states']
        energy_corrs = corrections['energies']
        all_energy_corrs = corrections['energy_corrections'] if 'energy_corrections' in corrections else None
        wfn_corrections = corrections['wavefunctions']
        if 'degenerate_states' in states:
            degenerate_states = states['degenerate_states']
        else:
            degenerate_states = None

        if 'degenerate_transformation' in corrections:
            degenerate_transf = corrections['degenerate_transformation']
        else:
            degenerate_transf = None

        if 'degenerate_energies' in corrections:
            degenerate_energies = corrections['degenerate_energies']
        else:
            degenerate_energies = None

        return cls(
            state_space,
            coupled_states,
            total_basis,
            energy_corrs,
            wfn_corrections,
            all_energy_corrections=all_energy_corrs,
            degenerate_states=degenerate_states,
            degenerate_transformation=degenerate_transf,
            degenerate_energies=degenerate_energies,
            **opts
        )

    @property
    def degenerate(self):
        """

        :return:
        :rtype:
        """
        return self.degenerate_transf is not None

    @property
    def energies(self):
        """

        :return:
        :rtype:
        """
        if self.degenerate:
            return self.degenerate_energies
        else:
            return np.sum(self.energy_corrs, axis=1)

    @property
    def order(self):
        """

        :return:
        :rtype:
        """
        return len(self.energy_corrs[0])

    def take_subspace(self, space):
        """
        Takes only those elements that are in space
        :param space:
        :type space:
        :return:
        :rtype:
        """

        new_states = self.states.find(space)
        # print("? =", new_states)
        return type(self)(
            self.states.take_subspace(new_states),
            self.coupled_states.take_states(space),
            self.total_basis,
            self.energy_corrs[new_states],
            [w[new_states, :] for w in self.wfn_corrections],
            # not sure what to do with all this...
            degenerate_states=self.degenerate_states,
            degenerate_transformation=self.degenerate_transf,
            degenerate_energies=self.degenerate_energies,
            logger=self.logger
        )

    @classmethod
    def create_coupling_matrix(cls, corrs,
                               states:BasisStateSpace, flat_total_space:BasisStateSpace,
                               nstates, order,
                               filters=None,
                               non_zero_cutoff=1.0e-14,
                               logger=None
                               ):
        """

        :param corrs:
        :type corrs:
        :param states:
        :type states:
        :param flat_total_space:
        :type flat_total_space:
        :param nstates:
        :type nstates:
        :param order:
        :type order:
        :param filters:
        :type filters:
        :param non_zero_cutoff:
        :type non_zero_cutoff:
        :return:
        :rtype:
        """

        # now we recompute reduced state spaces for use in results processing
        # and we also convert the correction vectors to sparse representations
        tci = flat_total_space.indices
        N = len(tci)

        is_transp = len(corrs) == order and len(corrs[0]) == nstates
        is_sparse = isinstance(corrs[0], SparseArray) if is_transp else isinstance(corrs, SparseArray)
        # nstates = len(all_corrs)

        corr_mats = [None] * (order + 1)
        corr_inds = [[] for i in range(nstates)]
        for o in range(order + 1):
            if filters is not None:
                # we use this to check that we're only
                # allowing transitions that our filters support
                nquanta_rules = [
                    v for k,v in filters.items() if sum(k) == o
                ]
            else:
                nquanta_rules = None
            non_zeros = []
            if is_sparse:
                if is_transp:
                    sp_vals, sp_inds = corrs[o].block_data
                    nonzi = np.where(np.abs(sp_vals) > non_zero_cutoff)[0]
                    sp_vals = sp_vals[nonzi,]
                    sp_inds = tuple(s[nonzi] for s in sp_inds)
                    corr_mats[o] = SparseArray.from_data(
                        (
                            sp_vals,
                            sp_inds
                        ),
                        shape=(nstates, N),
                        cache_block_data=False
                    )
                    ind_keys, ind_vals = nput.group_by(sp_inds[1], sp_inds[0])[0]
                    for i,v in zip(ind_keys, ind_vals):
                        corr_inds[i].append(v)
                else:
                    raise NotImplementedError("constructing final coupling matrix from (order, nstates, N) `SparseArray` not supported")
            else:
                initial_quanta = np.sum(states.excitations, axis=1)
                for i in range(nstates):
                    if is_transp:
                        nonzi = np.where(np.abs(corrs[o, i]) > non_zero_cutoff)[0]
                        vals = corrs[o, i][nonzi,]
                    else:
                        nonzi = np.where(np.abs(corrs[i, o]) > non_zero_cutoff)[0]
                        vals = corrs[i, o][nonzi,]

                    if len(nonzi) > 0:
                        # we attempt to filter out things that can't touch based on our filter rules
                        target_quanta = np.sum(flat_total_space.take_subspace(nonzi).excitations, axis=1)
                        if nquanta_rules is not None and len(nquanta_rules) > 0:
                            # from .Solver import PerturbationTheoryStateSpaceFilter
                            mask = None
                            for f in nquanta_rules:
                                # f:PerturbationTheoryStateSpaceFilter
                                for (filter_space, filter_rules) in f.prefilters:
                                    is_in = states.take_subspace([i]).intersection(filter_space)
                                    if len(is_in) > 0:
                                        q_diffs = target_quanta - initial_quanta[i]
                                        poss_diffs = np.unique([sum(x) for x in filter_rules])
                                        if mask is None:
                                            mask = np.isin(q_diffs, poss_diffs)
                                        else:
                                            mask = np.logical_or(mask, np.isin(q_diffs, poss_diffs))
                            if mask is not None:
                                nonzi = nonzi[mask]
                                vals = vals[mask]
                    else:
                        if logger is not None:
                            logger.log_print("No corrections for state {s} at order {o}",
                                s=states.excitations[i],
                                o=o
                            )

                    # and then we add the appropriate basis indices to the list of basis data
                    non_zeros.append(
                        (
                            vals,
                            np.column_stack([
                                np.full(len(nonzi), i),
                                nonzi
                            ])
                        )
                    )

                    corr_inds[i].append(tci[nonzi,])

                # now we build the full mat rep for this level of correction
                vals = np.concatenate([x[0] for x in non_zeros])
                inds = np.concatenate([x[1] for x in non_zeros], axis=0).T
                corr_mats[o] = SparseArray.from_data(
                    (
                        vals,
                        inds
                    ),
                    shape=(nstates, N),
                    cache_block_data=False
                )

        # now we build state reps from corr_inds
        for i, dat in enumerate(corr_inds): #TODO: this might break with pruning...I can't really be sure at this point
            spaces = []
            for substates in dat:
                _, upos = np.unique(substates, return_index=True)
                usubs = substates[np.sort(upos)]
                spaces.append(flat_total_space.take_states(usubs))
            corr_inds[i] = BasisMultiStateSpace(np.array(spaces, dtype=object))

        return corr_mats, corr_inds

    def prune(self, threshold=.1, in_place=False):
        """
        Returns corrections with couplings less than the given cutoff set to zero

        :param threshold:
        :type threshold:
        :return:
        :rtype:
        """
        if not in_place:
            import copy
            new = copy.copy(self)
            new.prune(threshold=threshold, in_place=False)
            # might need to work harder here...
            return new

        for o in range(1, self.order):
            sp_vals, sp_inds = self.wfn_corrections[o].block_data
            mask = abs(sp_vals) > threshold
            sp_vals = sp_vals[mask]
            sp_inds = tuple(s[mask] for s in sp_inds)
            self.wfn_corrections[o].block_vals = sp_vals
            self.wfn_corrections[o].block_inds = sp_inds
        return self

    def _take_subham(self, rep, inds):
        """
        Builds a subsampled version of a representation Hamiltonian
        to allow equations to be efficiently solved in subspaces.

        :param rep: representation matrix from which to pull the subspace
        :type rep: SparseArray
        :param inds: indices for the subspace
        :type inds: np.ndarray
        :return:
        :rtype:
        """
        ind_pairs = np.array(list(itertools.product(inds, inds))).T
        return np.reshape(rep[ind_pairs], (len(inds), len(inds)))
    def get_transformed_Hamiltonians(self, hams, deg_group=None):
        """

        :param corrs:
        :type corrs:
        :param deg_group:
        :type deg_group:
        :return:
        :rtype:
        """
        if deg_group is not None:
            subcorrs = self.take_subspace(deg_group)
            inds = self.total_basis.find(deg_group)
            subhams = [SparseArray.from_data(self._take_subham(H, inds)) for H in hams]
            # H_nd =[
            #     x.asarray() if isinstance(x, SparseArray) else x
            #     for x in subcorrs.operator_representation(subhams, subspace=deg_group)
            # ]
            H_nd = [
                x.asarray() if isinstance(x, SparseArray) else x
                for x in subcorrs.operator_representation(subhams, subspace=deg_group, logger_symbol="H")
            ]

        else:
            subhams = hams
            H_nd = [
                x.asarray() if isinstance(x, SparseArray) else x
                for x in self.operator_representation(subhams, logger_symbol="H", logger_conversion=UnitsData.convert("Hartrees", "Wavenumbers"))
            ]
        return H_nd
    def get_degenerate_rotation(self, deg_group, hams, label=None, zero_point_energy=None,
                                local_coupling_hamiltonian=None,
                                local_coupling_order=None):
        """

        :param deg_group:
        :type deg_group:
        :param corrs:
        :type corrs:
        :return:
        :rtype:
        """

        logger = self.logger
        # raise Exception(corrs.states.excitations, deg_group)
        with logger.block(tag="states" if label is None else label):
            logger.log_print(
                str(
                    self.states.take_states(deg_group).excitations
                ).splitlines()
            )
        subdegs = self.take_subspace(deg_group)

        # from McUtils.Scaffolding import JSONSerializer
        # import os
        # with open(os.path.expanduser("~/Desktop/wat6.json"), "w+") as woof:
        #     JSONSerializer().serialize(woof, subdegs)

        # H_nd = self.get_transformed_Hamiltonians(corrs, deg_group)
        # for h in H_nd[1:]:
        #     np.fill_diagonal(h, 0.)
        if local_coupling_order is None:
            local_coupling_order = -1
        if local_coupling_order < 0:
            local_coupling_order = len(hams) + local_coupling_order
        if local_coupling_order < len(hams) - 1:
            #TODO: calculate only the required terms at this order
            raise NotImplementedError("local mode couplings only provided at highest order")
        if local_coupling_hamiltonian is not None:
            hc = local_coupling_hamiltonian.get_representation_matrix(deg_group, deg_group,
                                                                      zero_element_warning=False,
                                                                      diagonal=False,
                                                                      # memory_constrained=self.memory_constrained
                                                                      )
        else:
            hc = None

        H_nd_corrs = subdegs.get_transformed_Hamiltonians(hams, deg_group=None)
        group_inds = self.states.find(deg_group)
        if hc is not None:
            H_nd_corrs[local_coupling_order] += hc.asarray()
        # zero_order_engs = self.energy_corrs[group_inds, 0]
        # raise Exception(
        #     (H_nd_corrs[0]-np.diag(np.full(len(group_inds), self.energy_corrs[0, 0])))*UnitsData.convert("Hartrees", "Wavenumbers"),
        #     (zero_order_engs-self.energy_corrs[0, 0])*UnitsData.convert("Hartrees", "Wavenumbers")
        # )
        # np.fill_diagonal(H_nd_corrs[0], zero_order_engs)

        # import McUtils.Plots as plt
        # plt.TensorPlot(np.array(H_nd)).show()
        H_nd = np.sum(H_nd_corrs, axis=0)
        if np.sum(H_nd) == 0:
            raise ValueError("No corrections from ", subdegs.wfn_corrections)
        #     raise Exception(deg_group.excitations,
        #                     self.states.take_states(deg_group).excitations,
        #                     # self.coupled_states.take_states(deg_group).excitations
        #                     )
        # overlaps = np.sum(subdegs.get_overlap_matrices(), axis=0)

        with logger.block(tag="non-degenerate Hamiltonian"):
            if zero_point_energy is None:
                zero_point_energy = self.energies[0]
            with np.printoptions(precision=self.nondeg_hamiltonian_precision, suppress=True):
                logger.log_print(
                    str(
                        (H_nd - np.diag(np.full(len(group_inds), zero_point_energy))) * UnitsData.convert("Hartrees", "Wavenumbers")
                    ).splitlines()
                )

        deg_engs, deg_transf = np.linalg.eigh(H_nd)

        ov_thresh = .5
        for i in range(len(deg_transf)):
            max_ov = np.max(deg_transf[:, i] ** 2)
            if max_ov < ov_thresh:  # there must be a single mode that has more than 50% of the initial state character?
                logger.log_print(
                    "    state {i} is more than 50% mixed",
                    i=i
                )
            #     raise PerturbationTheoryException("mode {} is has no contribution of greater than {}".format(
            #         i, ov_thresh
            #     ))

        # we pick the terms with the max contribution from each input state
        # and zero out the contributions so that two states can't map
        # to the same input state
        sort_transf = np.abs(deg_transf.copy())
        sorting = [-1] * len(deg_transf)
        for i in range(len(deg_transf)):
            o = np.argmax(sort_transf[i, :])
            sorting[i] = o
            sort_transf[:, o] = 0.  # np.zeros(len(sort_transf))

        with logger.block(tag='contributions'):
            logger.log_print(
                str(np.round(100 * (deg_transf ** 2)).astype(int)).splitlines()
            )

        logger.log_print('sorting: {s}', s=sorting)

        # sorting = np.argsort(sorting)

        #
        # # if len(sorting) != len(np.unique(sorting)):
        # #     raise PerturbationTheoryException("After diagonalizing can't distinguish modes...")
        deg_engs = deg_engs[sorting,]

        self.logger.log_print("degenerate energies {e}",
                              e=np.round(deg_engs * UnitsData.convert("Hartrees", "Wavenumbers")))

        deg_transf = deg_transf[:, sorting]

        return H_nd_corrs, deg_engs, deg_transf

    def get_degenerate_transformation(self, group, hams, gaussian_resonance_handling=False, label=None,
                                      zero_point_energy=None,
                                      local_coupling_hamiltonian=None,
                                      local_coupling_order=None
                                      ):
        # this will be built from a series of block-diagonal matrices
        # so we store the relevant values and indices to compose the SparseArray

        # we apply the degenerate PT on a group-by-group basis
        # by transforming the H reps into the non-degenerate basis
        # print(">", group.excitations)
        deg_inds = self.states.find(group, missing_val=-1)
        mask = deg_inds > -1
        deg_inds = deg_inds[mask]
        if not mask.all():
            bad = group.take_subspace(np.where(np.logical_not(mask))[0])
            self.logger.log_print(
                "WARNING: got degeneracy spec including states {} that are not in space of corrected states".format(
                    bad.excitations
                ))
        group = group.take_subspace(np.where(mask)[0])
        # print(">..", deg_inds, self.states.find(group, missing_val=-1))

        if len(deg_inds) == 1 or (
                gaussian_resonance_handling and np.max(np.sum(group.excitations, axis=1)) > 2):
            H_nd = deg_engs = deg_rot = None
        elif len(deg_inds) > 1:
            H_nd, deg_engs, deg_rot = self.get_degenerate_rotation(group, hams, label=label,
                                                                   zero_point_energy=zero_point_energy,
                                                                   local_coupling_hamiltonian=local_coupling_hamiltonian,
                                                                   local_coupling_order=local_coupling_order
                                                                   )
        else:
            H_nd = deg_engs = deg_rot = None
            # raise NotImplementedError("Not sure what to do when no states in degeneracy spec are in total space")

        return deg_inds, H_nd, deg_rot, deg_engs

    @staticmethod
    def default_state_filter(state, couplings, energy_cutoff=None, energies=None, basis=None, target_modes=None):
        """
        Excludes modes that differ in only one position, prioritizing states with fewer numbers of quanta
        (potentially add restrictions to high frequency modes...?)

        :param input_state:
        :type input_state:
        :param couplings:
        :type couplings:
        :return:
        :rtype:
        """

        if target_modes is None:
            target_modes = np.arange(len(state.excitations[0]))

        if energy_cutoff is not None:
            state_ind = basis.find(state)
            coupling_inds = basis.find(couplings)
            diff_mask = np.abs(energies[coupling_inds] - energies[state_ind]) > energy_cutoff
        else:
            exc_1 = state.excitations[0, target_modes]
            exc_2 = couplings.excitations[:, target_modes]
            diffs = exc_2 - exc_1[np.newaxis, :]
            diff_sums = np.sum(diffs != 0, axis=1)
            diff_mask = np.logical_or(
                np.sum(couplings.excitations, axis=1) == 0, # drop ground state
                diff_sums == 1 # find where changes are only in one position
            )
        # now drop these modes
        # print(couplings.excitations)
        if diff_mask.any():
            if diff_mask.all():
                couplings = None
            else:
                # print(">>>", state.excitations[0])
                # print(couplings.excitations)
                couplings = couplings.take_subspace(np.where(np.logical_not(diff_mask))[0])
                # print("===")
                # print(couplings.excitations)
        # print(couplings.excitations)
        return couplings

    def find_strong_couplings(self, threshold=.1, state_filter=None):
        """
        Finds positions in the expansion matrices where the couplings are too large

        :param threshold:
        :type threshold:
        :return:
        :rtype:
        """

        if state_filter is None:
            state_filter = self.default_state_filter

        order = self.order
        strong_couplings = {}
        for o in range(1, order):
            sp_vals, sp_inds = self.wfn_corrections[o].block_data
            nonzi = np.where(np.abs(sp_vals) > threshold)[0]
            sp_inds = tuple(s[nonzi] for s in sp_inds)
            ind_keys, ind_vals = nput.group_by(sp_inds[1], self.states.indices[sp_inds[0]])[0]
            # print(self.states.indices[sp_inds[0]])
            # print(tci[sp_inds[1]])
            for i, v in zip(ind_keys, ind_vals):
                _, upos = np.unique(v, return_index=True)
                usubs = v[np.sort(upos)]
                states = state_filter(self.states.take_states([i]), self.total_basis.take_subspace(usubs))
                if states is not None and len(states) > 0:
                    if i not in strong_couplings:
                        strong_couplings[i] = [None for _ in range(order)]
                    strong_couplings[i][o] = states

        return strong_couplings

    def format_strong_couplings_report(self, couplings=None, threshold=.1, int_fmt="{:>3.0f}", padding="{:<8}", join=True, use_excitations=True):
        if couplings is None:
            couplings = self.find_strong_couplings(threshold=threshold)

        list_fmt = " ".join(int_fmt for _ in range(self.total_basis.ndim))
        coupling_statements = []
        for i,v in sorted(couplings.items(), key=lambda k:k[0]):
            if use_excitations:
                i = self.states.take_states([i]).excitations[0]
                coupling_statements.append(padding.format("state:") + list_fmt.format(*i))
                # print(v)
                for n,l in enumerate(v):
                    if l is not None and len(l) > 0:
                        coupling_statements.extend(padding.format(" order {}".format(n) if j == 0 else "")+list_fmt.format(*e) for j,e in enumerate(l.excitations))
            else:
                coupling_statements.append(list_fmt.format(i))
                coupling_statements.append(padding + str(v.indices))

        return coupling_statements if not join else "\n".join(coupling_statements)

    def collapse_strong_couplings(self, sc:dict):
        """

        :param sc:
        :type sc:
        :return:
        :rtype:
        """
        new = {}
        for k,v in sc.items():
            s = None
            for s2 in v:
                if s2 is not None:
                    if s is None:
                        s = s2
                    else:
                        s = s.concatenate(s2)
            new[k] = s
        return new

    # def apply_martin_test(self, hams):
    #     zoos =
    #     (np.abs(H1_block) ** 4) / (diffs ** 3)

    @staticmethod
    def _fmt_operator_rep(full_ops, operator_symbol, conversion, real_fmt="{:>.8e}", padding_fmt='{:>16}'):
        tag_line = None
        rep_lines = None

        op_dim = None

        for (a, b, c), subrep in full_ops:
            if isinstance(subrep, SparseArray):
                subrep = subrep.asarray()
            elif isinstance(subrep, (int, float, np.integer, np.floating)):
                if subrep == 0:
                    if op_dim is None:
                        raise ValueError("was lazy and haven't filled operator dim yet...")
                    subrep = np.zeros(op_dim)
                else:
                    raise ValueError("don't know what to do with representation '{}'".format(subrep))

            if op_dim is None:
                op_dim = subrep.shape

            if conversion is not None:
                subrep = subrep * conversion

            subrep_lines = [
                " ".join(padding_fmt.format(real_fmt.format(e)) for e in line)
                for line in subrep
            ]
            line_len = len(subrep_lines[0])

            if rep_lines is None:
                rep_lines = subrep_lines
            else:
                rep_lines = [x + " " + y for x,y in zip(rep_lines, subrep_lines)]

            tag_fmt = "{:<" + str(line_len) + "}"
            base_tag=tag_fmt.format("<{a}|{A}({c})|{b}>".format(A=operator_symbol, a=a, c=c, b=b))
            if tag_line is None:
                tag_line = base_tag
            else:
                tag_line += " " + base_tag

        # we want to return a line list so the logger can add any prefixes it needs
        rep_lines.insert(0, tag_line)
        return rep_lines

    def operator_representation(self, operator_expansion,
                                order=None, subspace=None, contract=True,
                                logger_symbol="A",
                                logger_conversion=None
                                ):
        """
        Generates the representation of the operator in the basis of stored states

        :param operator_expansion: the expansion of the operator
        :type operator_expansion: Iterable[float] | Iterable[np.ndarray]
        :param order: the order of correction to go up to
        :type order: Iterable[float] | Iterable[np.ndarray]
        :param subspace: the subspace of terms in which the operator expansion is defined
        :type subspace: None | BasisStateSpace
        :return: the set of representation matrices for this operator
        :rtype: Iterable[np.ndarray]
        """

        mordor = self.order - 1
        if order is None:
            order = mordor
        if order > mordor:
            raise PerturbationTheoryException("{}: can't correct up to order {} when zero-order states were only corrected up to order {}".format(
                type(self).__name__,
                order,
                mordor
            ))
        order = order + 1 # so that we actually do get up to the request order after accounting for the zeros...
        if len(operator_expansion) < order:
            operator_expansion = list(operator_expansion) + [0]*(order - len(operator_expansion))

        # we stopped supporting indexing based on the total set of inds...
        if subspace is None:
            wfn_corrs = self.wfn_corrections[:order]
        else:
            # need to make the subspace good with the subspace in which the corrections are defined...
            subspace_sel = self.total_basis.find(subspace, check=True)
            wfn_corrs = []
            for k in range(order):
                wfn_corrs.append(self.wfn_corrections[k][:, subspace_sel])

        # generalizes the dot product so that we can use 0 as a special value...
        dot = _safe_dot
        logger = self.logger
        logger = None if logger is None or isinstance(logger, NullLogger) else logger

        # does the dirty work of acutally applying the rep...
        reps = [[] for _ in range(order)]
        full_ops = []
        for k in range(order):
            tags = []
            op = []
            # apply each thing up to requested order...
            for a in range(k+1): # if k == 2: a=0, a=1, a=2
                for b in range(k-a+1): # if k==2, a==0: b=0, b=1, b=2; a==1: b=0, b=1
                    c = k - (a + b) # a + b + c == k
                    rop = operator_expansion[c]
                    if isinstance(rop, (int, float, np.integer, np.floating)): # constant reps...
                        if rop != 0: # cheap easy check
                            subrep = rop * dot(wfn_corrs[a], wfn_corrs[b].T)
                            op.append(subrep)
                        else:
                            subrep = 0
                            op.append(0)
                    else:
                        subrep = dot(dot(wfn_corrs[a], rop), wfn_corrs[b].T)
                        op.append(subrep)

                    full_ops.append([
                        (a, b, c),
                        subrep
                    ])

            if contract:
                op = sum(op)
            reps[k] = op

        if logger is not None:
            logger.log_print(full_ops, logger_symbol, logger_conversion, message_prepper=self._fmt_operator_rep)

        return reps

    def get_overlap_matrices(self):
        """
        Returns the overlap matrices for the set of corrections
        at each order of correction

        :return:
        :rtype:
        """

        wat = []
        for k in range(2 + 1):
            ov = None
            for i in range(k + 1):
                c1 = self.wfn_corrections[i].asarray()
                c2 = self.wfn_corrections[k - i].asarray()
                if ov is None:
                    ov = np.dot(c1, c2.T)
                else:
                    ov += np.dot(c1, c2.T)
            wat.append(ov)

        return wat

    # def checkpoint_save(self, checkpoint:Checkpointer):
    #     """
    #     Writes correction arrays to checkpoint
    #
    #     :param checkpoint:
    #     :type checkpoint:
    #     :return:
    #     :rtype:
    #     """
    #     import gc, sys
    #
    #     # self.states = states
    #     # self.coupled_states = coupled_states
    #     # self.total_basis = total_basis
    #     # self.energy_corrs = energy_corrs
    #     # self.all_energy_corrs = all_energy_corrections
    #     # self.wfn_corrections = wfn_corrections
    #     # self.degenerate_states = degenerate_states
    #     # self.degenerate_transf = degenerate_transformation
    #     # self.degenerate_energies = degenerate_energies
    #     # self.logger = logger
    #     checkpoint["wfn_corrections"] = self.wfn_corrections
    #     print(">>>>", sys.getrefcount(self.wfn_corrections))
    #     self.wfn_corrections = None
    #     gc.collect()
    #
    # def checkpoint_reload(self, checkpoint:Checkpointer):
    #     self.wfn_corrections = checkpoint['wfn_corrections']
    #
    # def disk_backed(self):
    #     return self.chk_backer(self)
    #
    # class chk_backer:
    #     def __init__(self, parent):
    #         self.parent = parent
    #         self.chk = None
    #     def load_chk(self):
    #         import tempfile as tf
    #         target = tf.NamedTemporaryFile(suffix=".hdf5").name
    #         self.chk = Checkpointer.from_file(target)
    #     def unload_chk(self):
    #         import os
    #         os.remove(self.chk.checkpoint_file)
    #     def __enter__(self):
    #         self.load_chk()
    #         self.chk.__enter__()
    #         self.parent.checkpoint_save(self.chk)
    #     def __exit__(self, exc_type, exc_val, exc_tb):
    #         self.parent.checkpoint_reload(self.chk)
    #         self.chk.__exit__(exc_type, exc_val, exc_tb)
    #         self.unload_chk()

    def savez(self, file):
        raise NotImplementedError("old and wrong now")
        keys = dict(
            states=self.states,
            coupled_states=self.coupled_states,
            total_states=self.total_basis,
            energies=self.energy_corrs,
            wavefunctions=self.wfn_corrections
        )
        if self.degenerate_states is not None:
            keys['degenerate_states'] = self.degenerate_states
        if self.degenerate_transf is not None:
            keys['degenerate_transformation'] = self.degenerate_transf
        if self.degenerate_energies is not None:
            keys['degenerate_energies'] = self.degenerate_energies
        np.savez(file, **keys)
    @classmethod
    def loadz(cls, file):
        raise NotImplementedError("old and wrong now")
        keys = np.load(file)
        return cls.from_dicts(
            {
                "states":keys['states'],
                 "coupled_states":keys['coupled_states'],
                 "total_states":keys['total_states'],
                 "degenerate_states":keys['degenerate_states'] if 'degenerate_states' in keys else None
             },
            {
                "energies":keys['energies'],
                "wavefunctions":keys['wavefunctions'],
                "degenerate_transformation": keys['degenerate_transformation'] if 'degenerate_transformation' in keys else None,
                "degenerate_energies": keys['degenerate_energies'] if 'degenerate_energies' in keys else None
            },
            keys['hamiltonians']
        )

    def to_state(self, serializer=None):
        keys = dict(
            states=self.states,
            coupled_states=self.coupled_states,
            total_states=self.total_basis,
            energies=self.energy_corrs,
            wavefunctions=self.wfn_corrections,
            degenerate_states=self.degenerate_states,
            degenerate_transformations=self.degenerate_transf,
            degenerate_energies=self.degenerate_energies
        )
        return keys
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls.from_dicts(
            {
                "states": serializer.deserialize(data['states']),
                "coupled_states": serializer.deserialize(data['coupled_states']),
                "total_states": serializer.deserialize(data['coupled_states']),
                "degenerate_states": serializer.deserialize(data['degenerate_states']),
            },
            {
                "energies": data['energies'],
                "wavefunctions": data['wavefunctions'],
                "degenerate_transformation": data['degenerate_transformations'],
                "degenerate_energies": data['degenerate_energies']
            },
            data['hamiltonians'] # we probably want to ditch this for memory reasons...
        )


BasicAPTCorrections = collections.namedtuple("PTCorrections",
                                       ['initial_states', 'final_states', 'corrections']
                                       )

@dataclasses.dataclass
class AnalyticPerturbationTheoryCorrections:
    states: BasisStateSpace
    state_lists: 'list[tuple[np.ndarray, np.ndarray]]'
    _energies: np.ndarray = None
    _transition_moments: 'Iterable[np.ndarray]' = None
    _spectra: 'Iterable[DiscreteSpectrum]' = None
    _deperturbed_energies: np.ndarray = None
    _deperturbed_transition_moments: 'Iterable[np.ndarray]' = None
    _deperturbed_spectra: DiscreteSpectrum = None
    degenerate_states: 'Iterable[BasisStateSpace]' = None
    only_degenerate_terms: 'bool' = True
    _degenerate_hamiltonians: 'Iterable[np.ndarray]' = None
    _degenerate_coefficients: 'Iterable[np.ndarray]' = None
    _degenerate_state_list_transformations: 'Iterable[list[np.ndarray, np.ndarray]]' = None
    energy_corrections: BasicAPTCorrections = None
    transition_moment_corrections: 'Iterable[BasicAPTCorrections]' = None
    degenerate_hamiltonian_corrections: 'Iterable[BasicAPTCorrections]' = None
    operator_corrections: 'Iterable[BasicAPTCorrections]' = None
    _deperturbed_operator_values: 'Iterable[np.ndarray]' = None
    _operator_values: 'Iterable[np.ndarray]' = None
    operator_keys: 'Iterable[Any]' = None
    logger: 'Logger' = None

    _zpe_pos: int = None

    def get_zpe_pos(self) -> int:
        if self._zpe_pos is None:
            basis = self.states
            try:
                zpe_pos = basis.find([np.zeros(basis.ndim, dtype=int)])
            except IndexError:
                zpe_pos = 0  # do this better in the future...
            else:
                zpe_pos = zpe_pos[0]
            self._zpe_pos = zpe_pos
        return self._zpe_pos

    @property
    def energies(self) -> np.ndarray:
        if self._energies is None:
            if self.degenerate_states is None:
                self._energies = np.sum(self.energy_corrections, axis=0)
            else:
                self._deperturbed_energies = np.sum(self.energy_corrections, axis=0)
                self._energies, (self._degenerate_hamiltonians, self._degenerate_coefficients) = self.get_degenerate_transformations(
                    self.states, self._deperturbed_energies
                )
        return self._energies

    @property
    def deperturbed_energies(self) -> np.ndarray:
        if self._deperturbed_energies is None:
            self._deperturbed_energies = np.sum(self.energy_corrections, axis=0)
        return self._deperturbed_energies

    @classmethod
    def handle_degenerate_transformation(cls, degenerate_ham):
        deg_engs, deg_transf = np.linalg.eigh(degenerate_ham)

        # ov_thresh = .5
        # for i in range(len(deg_transf)):
        #     max_ov = np.max(deg_transf[:, i] ** 2)
        #     if max_ov < ov_thresh:  # there must be a single mode that has more than 50% of the initial state character?
        #         if logger is not None:
        #             logger.log_print(
        #                 "! state {i} is more than 50% mixed",
        #                 i=i
        #             )

        # we pick the terms with the max contribution from each input state
        # and zero out the contributions so that two states can't map
        # to the same input state
        sort_transf = np.abs(deg_transf.copy())
        sorting = [-1] * len(deg_transf)
        for i in range(len(deg_transf)):
            o = np.argmax(sort_transf[i, :])
            sorting[i] = o
            sort_transf[:, o] = 0.  # np.zeros(len(sort_transf))
        # print(">>>>", sorting)

        # self.logger.log_print('sorting: {s}', s=sorting)

        deg_engs = deg_engs[sorting,]
        deg_transf = deg_transf[:, sorting]
        return deg_engs, deg_transf

    def get_degenerate_transformations(self, basis, energies):
        energies = energies.copy()

        hams = []
        transf = []
        for block_num, (group, hcorr) in enumerate(
                zip(self.degenerate_states, self.degenerate_hamiltonian_corrections)
        ):
            ham = np.sum(hcorr, axis=0)
            # with self.logger.block(tag="Degenerate Block {n}", n=block_num + 1):
            #     self.logger.log_print("{s}",
            #                           s=group,
            #                           preformatter=lambda **kw: dict(
            #                               kw,
            #                               s=self.format_matrix(kw['s'])
            #                           ))
            e_pos = basis.find(group)
            if not self.only_degenerate_terms:
                ham = -ham  # only deg terms
            engs = energies[e_pos,]
            ham[np.diag_indices_from(ham)] = engs
            deg_engs, deg_mixing = self.handle_degenerate_transformation(ham)

            # with self.logger.block(tag='contributions'):
            #     self.logger.log_print(
            #         "{contribs}",
            #         contribs=deg_mixing,
            #         preformatter=lambda **kw: dict(
            #             kw,
            #             contribs=self.format_matrix(np.round(100 * (deg_mixing ** 2)))
            #         )
            #     )

            energies[e_pos] = deg_engs
            transf.append(deg_mixing)
            hams.append(ham)

        return energies, (hams, transf)

    @property
    def degenerate_hamiltonians(self):
        if self._degenerate_hamiltonians is None:
            e_base = self.energies # TODO: this is bad practice...
        return self._degenerate_hamiltonians
    @property
    def degenerate_coefficients(self):
        if self._degenerate_coefficients is None:
            e_base = self.energies # TODO: this is bad practice...
        return self._degenerate_coefficients

    def get_freqs(self):
        return self.energies - self.energies[self.get_zpe_pos()]

    def get_deperturbed_freqs(self):
        if self.degenerate_states is not None:
            return self.deperturbed_energies - self.deperturbed_energies[self.get_zpe_pos()]
        else:
            return self.get_freqs()

    @property
    def degenerate_transformation_pairs(self):
        if self._degenerate_state_list_transformations is None:
            self._degenerate_state_list_transformations = self._get_degenerate_tfs_mats()
        return self._degenerate_state_list_transformations
    def _get_degenerate_tfs_mats(self, logger=None):
        #TODO: add checks to ensure that our blocks are complete and we have proper unitary tfs at the end
        if logger is None:
            logger = self.logger
        logger = Logger.lookup(logger)
        all_degs = BasisStateSpace(
            HarmonicOscillatorProductBasis(len(self.state_lists[0][0][0])),
            np.concatenate(self.degenerate_states, axis=0)
        )
        deg_map_row = np.concatenate([
            [i] * len(g) for i, g in enumerate(self.degenerate_states)
        ])
        deg_map_col = np.concatenate([
            np.arange(len(g)) for i, g in enumerate(self.degenerate_states)
        ])
        for i, block in enumerate(self.degenerate_states):
            with logger.block(tag="Degenerate block {i}", i=i + 1):
                logger.log_print(
                    "{blocks}",
                    blocks=block,
                    preformatter=lambda **kw: dict(
                        kw,
                        blocks="\n".join(StateMaker.parse_state(e) for e in kw['blocks'])
                    )
                )

        tfs = []
        for block_idx, (init_states, final_states) in enumerate(self.state_lists):
            initial_space = BasisStateSpace(
                HarmonicOscillatorProductBasis(len(init_states[0])),
                init_states
            )
            final_space = BasisStateSpace(
                HarmonicOscillatorProductBasis(len(init_states[0])),
                final_states
            )
            init_pos = all_degs.find(initial_space, missing_val=-1)
            final_pos = all_degs.find(final_space, missing_val=-1)

            # subcorr is a n_init x n_final object, but we need to figure out the transformation to apply to each axis
            # to do so we find the appropriate transformation and insert it
            row_tf = np.eye(len(init_states))
            for n,i in enumerate(init_pos):
                if i != -1:
                    col_pos = deg_map_row[i]
                    row_pos = deg_map_col[i]
                    deg_block = self.degenerate_coefficients[col_pos]
                    block_idx = initial_space.find(self.degenerate_states[col_pos])
                    # print(row_pos, col_pos, col_tf.shape)
                    row_tf[block_idx, n] = deg_block[:, row_pos]

            nz_init_pos = [i for i in init_pos if i > -1]
            nz_init = [s for i, s in zip(init_pos, init_states) if i > -1]
            if len(nz_init_pos) > 0:
                with logger.block(tag="Initial states ({ix})",
                                  ix=[(deg_map_row[i],  deg_map_col[i]) for i in nz_init_pos]):
                    logger.log_print(
                        "{initials}",
                        initials=nz_init,
                        preformatter=lambda **kw: dict(
                            kw,
                            initials="\n".join(StateMaker.parse_state(e) for e in kw['initials'])
                        )
                    )
                    logger.log_print(
                        "{tf}",
                        tf=row_tf,
                        preformatter=lambda **kw:dict(kw, tf="\n".join(logger.prep_array(kw['tf'])))
                    )
            if np.sum(np.abs((row_tf @ row_tf.T) - np.eye(len(init_states))).flatten()) > 1e-3:
                raise ValueError("Non-unitary row tf, something wrong with initial state degs")

            col_tf = np.eye(len(final_states))
            for n,i in enumerate(final_pos):
                if i != -1:
                    col_pos = deg_map_row[i]
                    deg_block = self.degenerate_coefficients[col_pos]
                    block_idx = final_space.find(self.degenerate_states[col_pos])
                    row_pos = deg_map_col[i]
                    col_tf[block_idx, n] = deg_block[:, row_pos]
            nz_final_pos = [i for i in final_pos if i > -1]
            nz_final = [s for i, s in zip(final_pos, final_states) if i > -1]
            if len(nz_final_pos) > 0:
                with logger.block(tag="Final states ({ix})",
                                  ix=[(deg_map_row[i],  deg_map_col[i]) for i in nz_final_pos]):
                    logger.log_print(
                        "{finals}",
                        finals=nz_final,
                        preformatter=lambda **kw: dict(
                            kw,
                            finals="\n".join(StateMaker.parse_state(e) for e in kw['finals'])
                        )
                    )
                    logger.log_print(
                        "{tf}",
                        tf=col_tf,
                        preformatter=lambda **kw:dict(kw, tf="\n".join(logger.prep_array(kw['tf'])))
                    )
            if np.sum(np.abs((row_tf @ row_tf.T) - np.eye(len(init_states))).flatten()) > 1e-3:
                raise ValueError("Non-unitary col tf, something wrong with final state degs")

            tfs.append([row_tf, col_tf])
        return tfs
    def _apply_degs_to_corrs(self, corrs, logger=None):
        if logger is None:
            logger = self.logger
        logger = Logger.lookup(logger)


        has_subcorrs = not (isinstance(corrs[0], np.ndarray) and corrs[0].ndim == 2)
        if has_subcorrs:
            all_tms = [[] for _ in corrs]  # num axes
        else:
            all_tms = []
        tfs = self.degenerate_transformation_pairs

        for block_idx, (row_tf, col_tf) in enumerate(tfs):

            if has_subcorrs:
                for a, (storage, axis_moms) in enumerate(zip(all_tms, corrs)):
                    tf_corr = row_tf.T @ axis_moms[block_idx] @ col_tf
                    storage.append(tf_corr)
            else:
                all_tms.append(row_tf.T @ corrs[block_idx] @ col_tf)

        return all_tms

    @property
    def transition_moments(self):
        if self._transition_moments is None:
            logger = Logger.lookup(self.logger)
            null = NullLogger()
            if self.degenerate_states is None:
                self._transition_moments = self.deperturbed_transition_moments
            else:
                tmoms = self.deperturbed_transition_moments

                self._transition_moments = self._apply_degs_to_corrs(
                    tmoms,
                    logger = self.logger
                )

        return self._transition_moments

    @property
    def harmonic_transition_moments(self):
        return [
            [corr_block[0] for corr_block in axis]
            for axis in self.transition_moment_corrections
            # np.sum([np.sum(corr.corrections[n], axis=0) ** 2 for corr in self.transition_moment_corrections], axis=0)
            # for n in range(len(self.transition_moment_corrections[0]))
        ]

    @property
    def deperturbed_transition_moments(self):
        if self._deperturbed_transition_moments is None:
            self._deperturbed_transition_moments = [
                [np.sum(corr_block, axis=0) for corr_block in axis]
                for axis in self.transition_moment_corrections
                # np.sum([np.sum(corr.corrections[n], axis=0) ** 2 for corr in self.transition_moment_corrections], axis=0)
                # for n in range(len(self.transition_moment_corrections[0]))
            ]
        return self._deperturbed_transition_moments

    def get_spectra(self, energies, transition_moments):
        spectra = []
        for block_idx, (init_states, final_states) in enumerate(self.state_lists):
            block_specs = []
            init_engs = energies[self.states.find(init_states),]
            final_engs = energies[self.states.find(final_states),]
            for i, init in enumerate(init_states):
                submoms = np.array([
                    axis_moms[block_idx][i]
                    for axis_moms in transition_moments
                ]).T
                freqs = final_engs - init_engs[i]
                block_specs.append(DiscreteSpectrum.from_transition_moments(freqs, submoms))
            spectra.append(block_specs)
        return spectra

    @property
    def harmonic_spectra(self):
        return self.get_spectra(
            self.energy_corrections[0],
            self.harmonic_transition_moments
        )

    @property
    def deperturbed_spectra(self):
        if self._deperturbed_spectra is None:
            self._deperturbed_spectra = self.get_spectra(
                self.deperturbed_energies,
                self.deperturbed_transition_moments
            )
        return self._deperturbed_spectra

    @property
    def spectra(self):
        if self._spectra is None:
            if self.degenerate_states is None:
                self._spectra = self.deperturbed_spectra
            else:
                self._spectra = self.get_spectra(
                    self.energies,
                    self.transition_moments
                )
        return self._spectra

    @property
    def deperturbed_operator_values(self):
        if self._deperturbed_operator_values is None:
            self._deperturbed_operator_values = [
                [np.sum(corr_block, axis=0) for corr_block in op]
                for op in self.operator_corrections
            ]
        return self._deperturbed_operator_values

    @property
    def operator_values(self):
        if self._operator_values is None:
            if self.degenerate_states is None:
                self._operator_values = self.deperturbed_operator_values
            else:
                tmoms = self.deperturbed_operator_values

                self._operator_values = self._apply_degs_to_corrs(
                    tmoms,
                    logger=self.logger
                )

                # ops = self.deperturbed_operator_values
                # all_tms = [[] for _ in ops]  # num operators
                # for block_idx, (init_states, final_states) in enumerate(self.state_lists):
                #     for storage, axis_moms in zip(all_tms, ops):
                #         storage.append(
                #             self.apply_degenerate_transformations(
                #                 init_states, final_states,
                #                 axis_moms[block_idx]
                #             )
                #         )
                # self._operator_values = all_tms

        return self._operator_values



"""
Provides implementations of the Colbert-Miller DVR types defined in
http://xbeams.chem.yale.edu/~batista/v572/ColbertMiller.pdf
"""


import numpy as np
from .BaseDVR import BaseDVR

__all__ = [
    "CartesianDVR",
    "RingDVR",
    "RadialDVR",
    "PolarDVR",
]

class CartesianDVR(BaseDVR):
    """
    Provides the Colbert Miller DVR on the Cartesian [-inf, inf] range
    """

    def get_grid(self, domain=None, divs=None, **kw):
        """
        Provides the Colbert-Miller DVR grid for the [-inf, inf] range
        :param domain:
        :type domain:
        :param divs:
        :type divs:
        :param flavor:
        :type flavor:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        return domain[0] + (domain[1] - domain[0]) * np.arange(1, divs) / (divs + 1)

    def get_kinetic_energy(self, grid=None, mass=None, hb=1, **kwargs):

        dx = grid[1] - grid[0]  # recomputed here simply to decouple the calling from dvr_grid
        divs = len(grid)
        ke = np.empty((divs, divs))

        coeff = (hb ** 2) / (2 * mass * (dx ** 2))
        # compute the band values for the first row
        b_val_0 = coeff * (np.pi ** 2) / 3
        col_rng = np.arange(1, divs + 1)  # the column indices -- also what will be used for computing the off diagonal bands
        row_rng = np.arange(0, divs)  # the row indices -- computed once and sliced
        b_vals = coeff * ((-1) ** col_rng) * 2 / (col_rng ** 2)

        for i in range(divs):
            if i == 0:
                np.fill_diagonal(ke, b_val_0)
            else:
                col_inds = col_rng[i - 1:-1]  # +(i-1)
                row_inds = row_rng[:-i]
                ke[row_inds, col_inds] = b_vals[i - 1]
                ke[col_inds, row_inds] = b_vals[i - 1]

        return ke

    def real_momentum(self, grid=None, mass=None, hb=1, **kwargs):
        """
        Provides the real part of the momentum for the [-inf, inf] range
        :param grid:
        :type grid:
        :param hb:
        :type hb:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        divs = len(grid)
        dx = grid[1] - grid[0]
        p = np.zeros((divs, divs))

        col_rng = np.arange(1, divs + 1)
        row_rng = np.arange(0, divs)
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]  # +(i-1)
            row_inds = row_rng[:-i]
            val = hb / dx * (-1) ** (i) / i
            p[row_inds, col_inds] = val
            p[col_inds, row_inds] = -val

        return p

class RingDVR(BaseDVR):
    """
    Provides a DVR for working on the (0, 2Pi) range with periodicity from Colbert and Miller
    """

    def __init__(self, domain=None, **opts):
        if domain is None:
            domain = (0, 2*np.pi)
        super().__init__(domain=domain, **opts)

    def get_grid(self, domain=None, divs=None, **kw):
        """
        Provides the Colbert-Miller 1D grid for the [0, 2Pi] range
        :param domain:
        :type domain:
        :param divs:
        :type divs:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        if divs % 2 != 1:
            raise ValueError('number of DVR points must go as (2N + 1), i.e. it must be odd')

        return domain[0] + (domain[1] - domain[0]) * np.arange(1, divs + 1) / divs

    def get_kinetic_energy(self, grid=None, mass=1, hb=1, **kw):
        """
        Colbert-Miller kinetic energy for the [0, 2pi] range
        :param grid:
        :type grid:
        :param mass:
        :type mass:
        :param hb:
        :type hb:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        coeff = hb ** 2 / (2 * mass)
        divs = len(grid)
        n = (divs - 1) // 2
        ke = np.empty((divs, divs))

        diag_val = coeff * n * (n + 1) / 3
        np.fill_diagonal(ke, diag_val)

        col_rng = np.arange(1, divs + 1)  # the column indices -- also what will be used for computing the off diagonal bands
        row_rng = np.arange(0, divs)  # the row indices -- computed once and sliced
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]  # +(i-1)
            row_inds = row_rng[:-i]
            val = coeff * (-1) ** (i) * np.cos(i * np.pi / divs) / (2 * np.sin(i * np.pi / divs) ** 2)
            ke[row_inds, col_inds] = val
            ke[col_inds, row_inds] = val

        return ke

    def real_momentum(self, grid=None, hb=1, **kw):
        """
        Provides the real part of the momentum for the [0, 2pi] range
        :param grid:
        :type grid:
        :param hb:
        :type hb:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        # divs = len(grid)
        # p = np.zeros((divs, divs))
        # for j in range(1, divs+1):
        #     for jp in range(1, divs+1):
        #         i = jp - j
        #         if i == 0:
        #             p[j-1, jp-1] = 0
        #         else:
        #             val = hb / 2 * ((-1) ** i) / np.sin(i * np.pi / divs)
        #             p[j-1, jp-1] = val
                    # p[jp-1, j-1] = val
        # print("FIRST", p)

        coeff = hb / 2
        divs = len(grid)
        p = np.zeros((divs, divs))
        col_rng = np.arange(1, divs + 1)  # the column indices -- also what will be used for computing the off diagonal bands
        row_rng = np.arange(0, divs)  # the row indices -- computed once and sliced
        for i in range(1, divs): # loop over possible values of j - j'
            col_inds = col_rng[i-1:-1]  # +(i-1)
            row_inds = row_rng[:-i]
            val = coeff * ( (-1)**i ) / np.sin(i * np.pi / divs)
            p[row_inds, col_inds] = -val
            p[col_inds, row_inds] =  val

        return p

class PolarDVR(BaseDVR):
    """
    Provides a DVR for working on the (0, pi) range from Colbert and Miller
    """

    def __init__(self, domain=None, **opts):
        if domain is None:
            domain = (0, np.pi)
        super().__init__(domain=domain, **opts)

    def get_grid(self, domain=(0, np.pi), divs=None, **kwargs):
        """
        Provides the grid appropriate for the Colbert-Miller (0, Pi) range

        :param domain:
        :type domain:
        :param divs:
        :type divs:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if divs % 2 != 1:
            raise ValueError('number of DVR points must go as (2N + 1), i.e. it must be odd')

        return domain[0] + (domain[1] - domain[0]) * np.arange(1, divs) / divs

    def get_kinetic_energy(self, grid=None, mass=None, hb=1, **kwargs):
        """
        Colbert-Miller kinetic energy for the [0, pi] range
        :param grid:
        :type grid:
        :param mass:
        :type mass:
        :param hb:
        :type hb:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        coeff = hb ** 2 / (2 * mass)
        divs = len(grid)
        n = divs+1
        ke = np.empty((divs, divs))

        # we'll loop to fill to start
        ke[np.diag_indices_from(ke)] = (2*(n**2) + 1)/3 - 1/(np.sin(grid)^2)

        # then we start with constant diagonal bands like
        # in the previous cases but we also add on terms that
        # are constant along the _antidiagonal_ bands
        col_rng = np.arange(1, divs + 1)
        row_rng = np.arange(0, divs)
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]
            row_inds = row_rng[:-i]
            diag_val = 1 /(np.sin(np.pi*i/(2*n))**2)
            ke[row_inds, col_inds] = diag_val
            ke[col_inds, row_inds] = diag_val

        # we fill in the antidiagonal bands
        for i in range(2, 2*divs):
            #antidiag terms
            row_inds = row_rng[:i-1] if i <= divs else row_rng[i-divs-1:]
            antdiag_val = -1 /(np.sin(np.pi*i/(2*n))**2)
            ke[row_inds, np.flip(row_inds)] += antdiag_val
            # ke[flip_cols, row_inds] += val

        # finally we do the phases (hard to get in with the antidiagonals)
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]
            row_inds = row_rng[:-i]
            phase = (-1) ** (i)
            ke[row_inds, col_inds] *= phase
            ke[col_inds, row_inds] *= phase

        return coeff * ke

class RadialDVR(BaseDVR):
    """
    Provides a DVR for working on the (0, inf) range from Colbert and Miller
    """

    def get_grid(self, domain=(0, np.pi), divs=None, **kwargs):
        """
        Provides the grid appropriate for the Colbert-Miller (0, Pi) range

        :param domain:
        :type domain:
        :param divs:
        :type divs:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        return domain[0] + (domain[1] - domain[0]) * np.arange(1, divs) / (divs + 1)

    def get_kinetic_energy(self, grid=None, mass=None, hb=1, **kwargs):
        """
        Colbert-Miller kinetic energy for the (0, inf) range

        :param grid:
        :type grid:
        :param mass:
        :type mass:
        :param hb:
        :type hb:
        :param kw:
        :type kw:
        :return:
        :rtype:
        """

        dx = grid[1] - grid[0]
        coeff = (hb ** 2) / (2 * mass * (dx ** 2))
        divs = len(grid)
        ke = np.empty((divs, divs))

        # we'll loop to fill to start
        ke[np.diag_indices_from(ke)] = np.pi**2/3 - 1/np.arange(1, divs+1)

        # then we start with constant diagonal bands like
        # in the previous cases but we also add on terms that
        # are constant along the _antidiagonal_ bands
        col_rng = np.arange(1, divs + 1)
        row_rng = np.arange(0, divs)
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]
            row_inds = row_rng[:-i]
            diag_val = 2/(i**2)
            ke[row_inds, col_inds] = diag_val
            ke[col_inds, row_inds] = diag_val

        # we fill in the antidiagonal bands
        for i in range(2, 2*divs):
            #antidiag terms
            row_inds = row_rng[:i-1] if i <= divs else row_rng[i-divs-1:]
            antdiag_val = -2/(i**2)
            ke[row_inds, np.flip(row_inds)] += antdiag_val
            # ke[flip_cols, row_inds] += val

        # finally we do the phases (hard to get in with the antidiagonals)
        for i in range(1, divs):
            col_inds = col_rng[i - 1:-1]
            row_inds = row_rng[:-i]
            phase = (-1) ** (i)
            ke[row_inds, col_inds] *= phase
            ke[col_inds, row_inds] *= phase

        return coeff * ke


import os
import uuid
import numpy as np
import weakref
import tempfile as tf
import subprocess
import io
from McUtils.Jupyter import JHTML
from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput
import McUtils.Iterators as itut
from McUtils.ExternalPrograms import ExternalProgramRunner

__all__ = [
    "ezXML",
    "ezFCFInterface"
]

class ezXML:
    class ezTag(JHTML.XML.DeclarativeElement):
        all_remapped_attrs = weakref.WeakKeyDictionary()
        def __init__(self, *elems, flag=None, **attrs):
            preserved_attrs = {}
            if flag is not None:
                preserved_attrs["flag"] = self.sanitize_value(flag)
            remapped_attrs = {
                "attr-"+str(uuid.uuid4()): av
                for av in attrs.items()
            }
            self.all_remapped_attrs[self] = {k:v[0] for k,v in remapped_attrs.items()}
            super().__init__(
                *elems,
                **preserved_attrs,
                **{
                    k:self.sanitize_value(v[1]) for k,v in remapped_attrs.items()
                }
            )
        @classmethod
        def sanitize_value(cls, val):
            if isinstance(val, int):
                return str(val).lower()
            else:
                return str(val)
        def tostring(self, renormalize_padding=True):
            base = super().tostring(short_empty_elements=False, prettify=True)
            base = base.replace(
                "&#10;", "\n"
            )
            indent_marker = '---indent-marker---'
            pad = "\n" + indent_marker if renormalize_padding else ""
            for remapping in self.all_remapped_attrs.values():
                for k,v in remapping.items():
                    base = base.replace(k, pad+v)
            this_pad = pad if renormalize_padding else "\n"
            base = base.replace("></", ">" + this_pad + "</")
            # base = base.replace("\" />", "\"" + this_pad + "/>")
            if renormalize_padding:
                new_lines = []
                lines = base.splitlines()
                indent = ""
                for l in lines:
                    if l.strip().startswith("<") and not l.strip().startswith("</"):
                        indent = l[:l.index("<")] + "  "
                    elif l.startswith(indent_marker):
                        l = indent + l[len(indent_marker):]
                    new_lines.append(l)
                base = "\n".join(new_lines)
            return base

    class job_parameters(ezTag):  known_opts = {"temperature", "spectrum_intensity_threshold"}

    class do_not_excite_subspace(ezTag):  known_opts = {"size", "normal_modes"}

    class initial_state(ezTag):  known_opts = {}

    class target_state(ezTag):  known_opts = {}

    class the_only_initial_state(ezTag):  known_opts = {}

    class energy_thresholds(ezTag):  known_opts = {"units"}

    class print_franck_condon_matrices(ezTag):  known_opts = {"flag"}
    class print_normal_modes(ezTag):  known_opts = {"flag"}

    class parallel_approximation(ezTag):  known_opts = {"max_vibr_excitations_in_initial_el_state",
                                                        "max_vibr_excitations_in_target_el_state", "combination_bands",
                                                        "use_normal_coordinates_of_target_states"}

    class max_vibr_to_store(ezTag):  known_opts = {"target_el_state"}

    class single_excitation(ezTag):  known_opts = {"ini", "targ"}

    class dushinsky_rotations(ezTag):  known_opts = {"target_state", "max_vibr_excitations_in_initial_el_state",
                                                     "max_vibr_excitations_in_target_el_state"}

    class geometry(ezTag):
        known_opts = {"number_of_atoms", "linear", "units", "text"}
        def __init__(self,
                     *,
                     number_of_atoms,
                     text,
                     linear="false",
                     units="angstr",
                     **opts
        ):
            super().__init__(
                number_of_atoms=number_of_atoms,
                linear=linear,
                units=units,
                **opts,
                text=text
            )

    class manual_atoms_reordering(ezTag):  known_opts = {"new_order"}

    class manual_coordinates_transformation(ezTag):
        known_opts = {
            "shift_along_x", "shift_along_y", "shift_along_z",
            "rotate_around_x", "rotate_around_y", "rotate_around_z"
        }

    class normal_modes(ezTag):
        known_opts = {"if_mass_weighted", "text", "atoms"}
        def __init__(self,
                     *,
                     atoms,
                     if_mass_weighted,
                     text,
                     **opts
        ):
            super().__init__(
                atoms=atoms,
                if_mass_weighted=if_mass_weighted,
                **opts,
                text=text
            )

    class frequencies(ezTag):  known_opts = {"text"}

    class excitation_energy(ezTag):  known_opts = {"units"}

    class manual_normal_modes_reordering(ezTag):  known_opts = {"new_order"}

    class input(ezTag):  known_opts = {"job"}


class ezFCFInterface:
    def __init__(self, atoms, gs_nms, es_nms, excitations,
                 masses=None,
                 ground_states=None,
                 include_rotation=True, rotation_order='gs',
                 rotation_blocks=None,
                 rotation_center=None, logger=None, # ignored
                 mode_reordering=None,
                 rotation_method='duschinsky',
                 mass_weight=False,
                 dimensionless=False,
                 always_run_parallel=True,
                 print_all=True,
                 embed=False):
        self.atoms, self.atom_map = self.prep_masses(atoms, masses)
        self.excitation = excitations
        self.ground_states = ground_states
        if embed:
            raise NotImplementedError("need to add embedding to this branch too...")
        if rotation_method is not None and rotation_method != 'duschinsky':
            raise ValueError(f"ezFCF only supports Duschinsky rotations (got {rotation_method})")
        if dimensionless:
            raise ValueError("ezFCF doesn't support dimensionless normal modes")
        if mass_weight:
            gs_nms = gs_nms.make_mass_weighted()
            es_nms = es_nms.make_mass_weighted()
        self.gs_nms = gs_nms
        self.es_nms = es_nms
        self.include_rotation = include_rotation
        self.rotation_order = rotation_order
        self.rotation_blocks = rotation_blocks
        self.mode_reordering = mode_reordering
        self.always_run_parallel = always_run_parallel
        self.print_all = print_all

    @classmethod
    def _check_listable(cls, excitations):
        try:
            return all(
                isinstance(exc, np.ndarray) or all(
                    nput.is_numeric(e) or all(nput.is_numeric(ee) for ee in e)
                    for e in exc
                )
                for exc in excitations
            )
        except TypeError:
            return False


    def format(self, job_type="harmonic_pes", temperature=0, spectrum_intensity_threshold=1e-8):
        els = [
        ]
        if self.print_all:
            els.append(ezXML.print_normal_modes(flag="true"))
            els.append(ezXML.print_normal_modes(flag="true"))

        els.append(
            ezXML.job_parameters(
                temperature=temperature,
                spectrum_intensity_threshold=spectrum_intensity_threshold
            )
        )

        excitation_options, excitation_elems = self.prep_excitations(self.es_nms, self.ground_states, self.excitation)
        include_rotation = self.include_rotation
        if not include_rotation or self.always_run_parallel:
            els.append(
                self.prep_parallel(
                    *excitation_elems,
                    **excitation_options
                )
            )
        if include_rotation:
            els.append(
                self.prep_duschinsky(
                    *excitation_elems,
                    **excitation_options
                )
            )

        mode_reordering = self.mode_reordering
        return ezXML.input(
            *els,
            self.prep_initial_state(self.atoms, self.gs_nms),
            self.prep_target_state(self.atoms, self.es_nms, mode_reordering=mode_reordering),
            job=job_type
        )

    def prep_masses(self, atoms, masses, units=None):
        if masses is None:
            masses = [AtomData[a]["Mass"] for a in atoms]
        if units is None:
            if masses[0] > 500:
                units = "ElectronMass"
            else:
                units = "AtomicMassUnits"
        masses = np.array(masses) * UnitsData.convert(units, "AtomicMassUnits")

        if atoms is None:
            atoms = [""] * len(masses)
            rem = np.arange(len(masses))
            for key, val in AtomData.data.items():
                diffs = val["Mass"] - masses[rem,]
                matches = np.where(abs(diffs) < .05)
                if len(matches) > 0 and len(matches[0]) > 0:
                    match_pos = rem[matches]
                    for i in match_pos:
                        atoms[i] = val["Symbol"]
                    rem = np.delete(rem, match_pos[0])
                    if len(rem) == 0:
                        break
            else:
                for i in rem: atoms[i] = "X"

        atom_map = {}
        new_atoms = []
        ndig = int(1 + np.floor(np.log10(len(atoms))))
        for i,(atom, mass) in enumerate(zip(atoms, masses)):
            if abs(mass - atom_map.get(atom, mass)) > 1e-4:
                atom = atom + f"00{i:0>{ndig}d}"
                while atom in atom_map:
                    atom = atom + f"0{i}"
            atom_map[atom] = mass
            new_atoms.append(atom)
        return new_atoms, atom_map

    def format_masses_file(self, atom_map):
        return "<masses>\n" + "\n".join(f"  <{at}> {mass} </{at}>" for at, mass in atom_map.items()) + "\n</masses>"

    class _write_dir:
        def __init__(self, dir=None, dir_prefix=None, dir_suffix=None, delete=True):
            self.dir = dir
            self._temp_dirs = []
            self.delete = delete
            self.opts = {'prefix':dir_prefix, 'suffix':dir_suffix}
        def __enter__(self):
            if self.dir is None:
                td = tf.TemporaryDirectory(**self.opts)
                self._temp_dirs.append(td)
                return td.__enter__()
            else:
                return self.dir
        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.dir is None:
                td = self._temp_dirs.pop()
                td.__exit__(exc_type, exc_val, exc_tb)

    class ezFCFRunner(ExternalProgramRunner):
        default_opts = dict(prefix='ezFCF-', suffix='.xml')
        def __init__(self, job, binary, **opts):
            self.job = job
            super().__init__(binary, **opts)
        def prep_dir(self, dir):
            with open(os.path.join(dir, 'atomicMasses.xml'), 'w+') as mass_file:
                mass_file.write(self.job.format_masses_file(self.job.atom_map) + "\n\n")

        text_file_extensions = ExternalProgramRunner.text_file_extensions + ['.spectrum_parallel', '.spectrum_duschinsky']
    def run(self, ezFCF_binary,
            dir=None, dir_prefix=None, dir_suffix=None, mode='w',
            prefix='ezFCF-', suffix='.xml', delete=True,
            raise_errors=True,
            **job_opts):
        return self.ezFCFRunner(
            self,
            ezFCF_binary,
            dir=dir, dir_prefix=dir_prefix, dir_suffix=dir_suffix, mode=mode,
            prefix=prefix, suffix=suffix, delete=delete,
            raise_errors=raise_errors
        ).run(
            self.format(**job_opts).tostring() + "\n\n",
            return_auxiliary_files={"parallel": "{name}.spectrum_parallel", "duschinsky": '{name}.spectrum_dushinsky'}
        )

        # results = {}
        # with self._write_dir(dir=dir, dir_prefix=dir_prefix, dir_suffix=dir_suffix) as dir:
        #     with tf.NamedTemporaryFile(dir=dir, mode=mode, prefix=prefix, suffix=suffix, delete=False) as inp:
        #         results['input_file'] = self.format(**job_opts).tostring() + "\n\n"
        #         inp.write(results['input_file'])
        #     try:
        #         if os.path.isfile(ezFCF_binary):
        #             ezFCF_binary = os.path.abspath(ezFCF_binary)
        #         res = subprocess.run([ezFCF_binary, inp.name], capture_output=True, cwd=dir)
        #         results['process_output'] = res
        #         for key,ext in {"parallel":".spectrum_parallel", "duschinsky":'.spectrum_dushinsky'}.items():
        #             test = inp.name + ext
        #             if os.path.isfile(test):
        #                 with open(test) as strm:
        #                     results[key] = strm.read()
        #         err = res.stderr.decode().strip()
        #         if raise_errors and len(err) > 0:
        #             raise IOError(err)
        #     finally:
        #         if delete:
        #             try:
        #                 os.remove(inp.name)
        #             except OSError:
        #                 pass
        #
        # if not delete:
        #     return inp, results
        # else:
        #     return results
        # # if dir is not None:

    _exc_aliases = {
        'max_quanta':'max_vibr_excitations_in_target_el_state'
    }
    _gsstate_aliases = {
        'max_quanta':'max_vibr_excitations_in_initial_el_state'
    }
    @classmethod
    def canonicalize_excitation_options(cls, nms, threshold=None, fixed_modes=None, ground_states=None, **opts):
        base_opts = {cls._exc_aliases.get(k, k):v for k,v in opts.items()}
        elems = []

        gs_threshold = None
        if ground_states is not None:
            if isinstance(ground_states, dict):
                if 'threshold' in ground_states:
                    ground_states = ground_states.copy()
                    gs_threshold = ground_states['threshold']
                    del ground_states['threshold']
                base_opts.update(
                    {
                        cls._gsstate_aliases.get(k, k): v for k, v in ground_states.items()
                        if k not in {'threshold'}
                    }
                )
            else:
                elems.append(
                    ezXML.the_only_initial_state(cls.prep_state_str(nms, ground_states))
                )

        if fixed_modes is not None:
            if nput.is_numeric(fixed_modes): fixed_modes = [fixed_modes]
            elems.append(
                ezXML.do_not_excite_subspace(
                    size=len(fixed_modes),
                    norm_modes=" ".join(str(n) for n in fixed_modes)
                )
            )
        if threshold is not None or gs_threshold is not None:
            subelems = []
            if gs_threshold is not None:
                subelems.append(
                    ezXML.initial_state(
                        gs_threshold * UnitsData.convert("Hartrees", "Wavenumbers"),
                        units='cm-1'
                    )
                )
            if threshold is not None:
                subelems.append(
                    ezXML.target_state(
                        threshold * UnitsData.convert("Hartrees", "Wavenumbers"),
                        units='cm-1'
                    )
                )
            elems.append(
                ezXML.energy_thresholds(*subelems)
            )
        return base_opts, elems


    @classmethod
    def prep_excitations(cls, nms, ground_states, excitations):
        """
        Dispatcher to get appropriate state spaces
        :param excitations:
        :param check:
        :return:
        """

        opts = {}
        elems = []
        states = []

        try:
            exc0 = excitations[0][0]
        except (IndexError, KeyError, TypeError):
            refilter = True
        else:
            refilter = (
                    not isinstance(exc0, int)
                    or any(len(e) < len(nms.freqs) for e in excitations)
            )

        if refilter:
            if isinstance(excitations, dict):
                new_opts, new_elems = cls.canonicalize_excitation_options(nms, **excitations)
                opts.update(new_opts)
                elems.extend(new_elems)
            elif isinstance(excitations, (int, np.integer)) or isinstance(excitations[0], (int, np.integer)):
                opts.update({'max_vibr_excitations_in_target_el_state': excitations})
            elif cls._check_listable(excitations):
                from ..BasisReps import StateMaker
                state = StateMaker(len(nms.freqs), mode='normal')
                states.extend(state(*exc) for exc in states)
            else:
                raise ValueError("can't build a state space from {}".format(excitations))
        else:
            states.extend(excitations)

        gs_states = []
        if ground_states is not None:
            # gs_states = []

            try:
                exc0 = excitations[0][0]
            except (IndexError, KeyError, TypeError):
                refilter = True
            else:
                refilter = (
                        not isinstance(exc0, int)
                        or any(len(e) < nms.freqs for e in excitations)
                )

            if refilter:
                if isinstance(excitations, dict):
                    new_opts, new_elems = cls.canonicalize_excitation_options(excitations)
                    opts.update(new_opts)
                    elems.extend(new_elems)
                elif isinstance(excitations, (int, np.integer)) or isinstance(excitations[0], (int, np.integer)):
                    opts.update({'max_vibr_excitations_in_target_el_state': excitations})
                elif cls._check_listable(excitations):
                    from ..BasisReps import StateMaker
                    state = StateMaker(len(nms.freqs), mode='normal')
                    gs_states.extend(state(*exc) for exc in states)
                else:
                    raise ValueError("can't build a state space from {}".format(excitations))
            else:
                gs_states.extend(excitations)

            if len(gs_states) == 1:
                elems.append(
                    ezXML.the_only_initial_state(cls.prep_state_str(nms, ground_states))
                )
            # elif len(gs_states) > 2:
            #     raise ValueError("ezFCF doesn't support state-by-state corrections, need to find a new approach")
        else:
            opts['max_vibr_excitations_in_initial_el_state'] = 0

        if len(states) > 0:
            if len(gs_states) == 0:
                gs_states = [[0]*len(nms.freqs)]
            starts = [
                cls.prep_state_str(nms, s) for s in gs_states
            ]
            ends = [
                cls.prep_state_str(nms, s) for s in states
            ]
            for i in starts:
                for j in ends:
                    elems.append(
                        ezXML.single_excitation(
                            ini=i,
                            targ=j
                        )
                    )
            # raise ValueError("ezFCF doesn't support state-by-state corrections, need to find a new approach")
        return opts, elems

    @classmethod
    def prep_state_str(cls, nms, state):
        if isinstance(state, str):
            return state
        else:
            if not len(state) == len(nms.freqs) or not nput.is_numeric(state[0]):
                from ..BasisReps import StateMaker
                state = StateMaker(len(nms.freqs), mode='normal')(*state)
            nzp = np.nonzero(state)
            if len(nzp) == 0 or len(nzp[0]) == 0: return "0"
            return ",".join(f"{i}v{state[i]}" for i in nzp[0])

    def prep_parallel(self, *elems, rotation_order='gs',
                      max_vibr_excitations_in_initial_el_state=0,
                      max_vibr_excitations_in_target_el_state=0,
                      **etc):
        elems = list(elems)
        opts = dict(
            dict(use_normal_coordinates_of_target_states=rotation_order == 'gs'),
            **etc
        )
        if self.print_all:
            elems.append(ezXML.print_franck_condon_matrices(flag=True))
            elems.append(ezXML.print_franck_condon_matrices(flag=True))
        return ezXML.parallel_approximation(
            *elems,
            max_vibr_excitations_in_initial_el_state=max_vibr_excitations_in_initial_el_state,
            max_vibr_excitations_in_target_el_state=max_vibr_excitations_in_target_el_state,
            **opts
        )

    def prep_duschinsky(self, *elems, rotation_order='gs',
                        max_vibr_excitations_in_initial_el_state=0,
                        max_vibr_excitations_in_target_el_state=0,
                        **etc):
        elems = list(elems)
        opts = dict(
            dict(target_state=0 if rotation_order == 'es' else 1),
            **etc
        )
        if self.print_all:
            elems.append(ezXML.print_franck_condon_matrices(flag=True))
        return ezXML.dushinsky_rotations(
            *elems,
            max_vibr_excitations_in_initial_el_state=max_vibr_excitations_in_initial_el_state,
            max_vibr_excitations_in_target_el_state=max_vibr_excitations_in_target_el_state,
            **opts
        )

    @classmethod
    def prep_initial_state(cls, atoms, nms, autoembed=False, excitation_energy_units="cm-1"):
        els = [
            # ezXML.excitation_energy(units=excitation_energy_units),
        ]
        if not autoembed:
            els.append(
                ezXML.manual_coordinates_transformation(
                    shift_along_x=0, shift_along_y=0, shift_along_z=0,
                    rotate_around_x=0, rotate_around_y=0, rotate_around_z=0
                )
            )
        return cls.prep_state(ezXML.initial_state,
                              atoms, nms,
                              *els#excitation_energy_units=excitation_energy_units
                              )

    @classmethod
    def prep_target_state(cls, atoms, nms, autoembed=False, mode_reordering=None, excitation_energy_units="cm-1"):
        """

        :param atoms:
        :param nms:
        :param excitation_energy_units:
        :return:
        """
        els = [
            ezXML.excitation_energy(units=excitation_energy_units),
        ]
        if not autoembed:
            els.append(
                ezXML.manual_coordinates_transformation(
                    shift_along_x=0, shift_along_y=0, shift_along_z=0,
                    rotate_around_x=0, rotate_around_y=0, rotate_around_z=0
                )
            )
        if mode_reordering is not None:
            els.append(
                ezXML.manual_normal_modes_reordering(
                    new_order=" ".join(str(int(x)) for x in mode_reordering)
                )
            )
        return cls.prep_state(
            ezXML.target_state,
            atoms, nms,
            *els,
        )

    @classmethod
    def prep_state(cls, tag, atoms, nms, *elems):
        return tag(
            *elems,
            cls.prep_geometry(atoms, nms),
            cls.prep_normal_modes(atoms, nms),
            cls.prep_frequencies(nms)
        )

    @classmethod
    def prep_geometry(cls, atoms, nms, linear=False, units="BohrRadius"):
        return ezXML.geometry(
            number_of_atoms=str(len(atoms)),
            linear=str(linear).lower(),
            units="angstr",
            text="\n" + "\n".join(
                f"{a:<8} {xyz[0]: >12.6f}  {xyz[1]: >12.6f}  {xyz[2]: >12.6f}"
                for a,xyz in zip(
                    atoms,
                    nms.remove_mass_weighting().origin * UnitsData.convert(units, "Angstroms")
                )
            ) + "\n"
        )

    @classmethod
    def format_modes_block(cls, modes):
        splits = itut.chunked(modes.T, 3)
        lines = []
        for block in splits:
            block = np.concatenate([b.reshape(-1, 3) for b in block], axis=1)
            block = [
                ['{:>10.6f}'.format(b) for b in l] + [""] * (9 - len(l))
                for l in block
            ]
            block_lines = [
                (
                    "   " + (" ".join(l[:3]))
                    + "   " + (" ".join(l[3:6]))
                    + "   " + (" ".join(l[6:9]))
                ).rstrip()
                for l in block
            ]
            lines.extend(block_lines)
            lines.append("")

        return "\n".join(lines)

    @classmethod
    def parse_modes_block(cls, modes_str, num_atoms=None):
        # blocks = [b.strip() for b in modes_str.splitlines()]
        # block_groups = itut.split(blocks, lambda l,_:len(l.strip()) == 0)
        modes = []
        if num_atoms is not None:
            base_array = np.loadtxt(io.StringIO(modes_str.replace("-", " -").replace("\n", " "))).flatten()
            nchunks = len(base_array) // (3 * 3 * num_atoms)
            blocks = np.array_split(base_array, nchunks)
            for block in blocks:
                modes.extend(b.T.flatten()[:, np.newaxis] for b in itut.chunked(block.reshape(num_atoms, -1).T, 3))
        else:
            blocks = modes_str.split("\n\n")
            for block in blocks:
                if len(block.strip()) > 0:
                    block = np.loadtxt(io.StringIO(block.replace("-", " -")), dtype=float)
                    modes.extend(b.T.flatten()[:, np.newaxis] for b in itut.chunked(block.T, 3))
        return np.concatenate(modes, axis=1)

    @classmethod
    def prep_normal_modes(cls, atoms, nms):
        modes = nms.modes_by_coords / UnitsData.convert("BohrRadius", "Angstroms")
        if nms.mass_weighted:
            modes = modes / np.sqrt(UnitsData.convert("ElectronMass", "AtomicMassUnits"))
        return ezXML.normal_modes(
            atoms=" ".join(atoms),
            text="\n"+cls.format_modes_block(modes) + "\n",
            if_mass_weighted=str(nms.mass_weighted).lower()
        )

    @classmethod
    def format_freqs_block(cls, freqs):
        splits = itut.chunked(freqs, 3)
        lines = []
        for l in splits:
            l = [" " + '{:>10.2f}'.format(b) for b in l] + [""] * (3 - len(l))
            lines.append("".join(l).strip())

        return "\n".join(lines)
    @classmethod
    def parse_freqs_block(cls, freqs):
        return np.loadtxt(io.StringIO(freqs.replace("\n", " ")))

    @classmethod
    def prep_frequencies(cls, nms):
        return ezXML.frequencies(
            text="\n"+cls.format_freqs_block(nms.freqs * UnitsData.convert("Hartrees", "Wavenumbers")) + "\n"
        )

    @classmethod
    def parse_state(self, state_xml):
        from ..Molecools import Molecule

        if isinstance(state_xml, str):
            state_xml = JHTML.XML.parse(state_xml)

        nms_block = state_xml.find("normal_modes")
        try:
            mw = nms_block["if_mass_weighted"]
        except AttributeError:
            mw = False
        else:
            mw = mw in {"true", "y"}
        ats = nms_block["atoms"].split()
        raw_modes = ezFCFInterface.parse_modes_block(
            nms_block["text"],
            num_atoms=len(ats)
        ) #/ UnitsData.convert("Angstroms", "BohrRadius")
        if mw:
            mode_norms = np.diag(raw_modes.T @ raw_modes)
            scaling = mode_norms[0]
            if scaling > 10:
                scaling = UnitsData.convert("Angstroms", "BohrRadius") * np.sqrt(
                    UnitsData.convert("AtomicMassUnits", "ElectronMass")
                )
            elif scaling > 2:
                scaling = UnitsData.convert("Angstroms", "BohrRadius")
            else:
                scaling = 1
            raw_modes = raw_modes * scaling

            masses = 1 / np.sqrt(
                    np.array([AtomData[s, "Mass"] for s in ats])
                    * UnitsData.convert("AtomicMassUnits", "ElectronMass")
            )
            masses = np.repeat(masses, 3)
            modes = np.diag(masses) @ raw_modes
            inverse = raw_modes.T @ np.diag(1/masses)
            # modes = raw_modes
        else:
            # TODO: shoud
            inverse = raw_modes.T
            modes = np.linalg.pinv(inverse)
        freqs = ezFCFInterface.parse_freqs_block(
            state_xml.find("frequencies")["text"]
        ) * UnitsData.convert("Wavenumbers", "Hartrees")

        return Molecule.from_string(
            state_xml.find("geometry")["text"],
            "xyz",
            units="Angstroms",
            normal_modes={
                "matrix": modes,
                "inverse": inverse,
                "freqs": freqs
            }
        )

    @classmethod
    def parse_fc_model(cls, input_xml, **opts):
        from .FCFs import FranckCondonModel

        if os.path.isfile(input_xml):
            with open(input_xml) as input_file:
                input_xml = input_file.read()

        if isinstance(input_xml, str):
            input_xml = JHTML.XML.parse(input_xml)

        gs = cls.parse_state(input_xml.find('initial_state'))
        es = cls.parse_state(input_xml.find('target_state'))

        include_rotation = input_xml.find('duschinsky_rotations') is not None
        mode_reordering = input_xml.find('target_state').find("manual_normal_modes_reordering")
        if mode_reordering is not None:
            mode_reordering = [int(s) for s in mode_reordering['new_order'].split()]

        return FranckCondonModel.from_mols(
            gs, es,
            include_rotation=include_rotation,
            mode_reordering=mode_reordering,
            remove_transrot=False,
            **opts
        )



"""
Developed to decoupled molecules and vibrations, intended to become the
replacement for the already developed `MolecularNormalModes` / `MolecularVibrations`
"""

__all__ = []
from .MixtureModes import *; from .MixtureModes import __all__ as exposed
__all__ += exposed
from .NormalModes import *; from .NormalModes import __all__ as exposed
__all__ += exposed
from .ObliqueModes import *; from .ObliqueModes import __all__ as exposed
__all__ += exposed
from .LocalizedModes import *; from .LocalizedModes import __all__ as exposed
__all__ += exposed
from .Redundant import *; from .Redundant import __all__ as exposed
__all__ += exposed


"""
Provides support for handling modes that arise from
"""

import numpy as np
import McUtils.Numputils as nput
from McUtils.Coordinerds import CoordinateSystem, CartesianCoordinateSystem3D, InternalCoordinateSystem

# from .MoleculeInterface import AbstractMolecule
# from .Transformations import MolecularTransformation

__all__ = [
    "MixtureModes"
]
class MixtureModes(CoordinateSystem):
    """
    A `McUtils.Coordinerds.CoordinateSystem` object that expresses coordinates as
    a rotation on some base set of coordinates with some associated frequencies.
    """
    name="MixtureModes"
    def __init__(self,
                 basis,
                 coeffs,
                 freqs=None,
                 origin=None,
                 masses=None,
                 inverse=None,
                 mass_weighted=False,
                 frequency_scaled=False,
                 g_matrix=None,
                 name=None,
                 ):
        if (
                isinstance(coeffs, np.ndarray) or
                (not nput.is_numeric(coeffs[0]) and nput.is_numeric(coeffs[0][0]))
        ):
            coeffs = [coeffs]
        coeffs = [np.asanyarray(c) for c in coeffs]
        full_coeffs = coeffs
        coeffs = coeffs[0]

        super().__init__(
            matrix=coeffs,
            inverse=inverse,
            name=self.name if name is None else name,
            basis=basis,
            dimension=(coeffs.shape[1],),
            origin=origin
        )
        self.freqs = freqs
        self.masses = masses
        self.mass_weighted = mass_weighted
        self.frequency_scaled = frequency_scaled
        self.g_matrix = g_matrix
        self._extended_coeffs = full_coeffs
        self._inverse_coeffs = None

    @classmethod
    def prep_modes(cls, modes):
        if isinstance(modes, cls) or all(hasattr(modes, b) for b in ['basis', 'matrix', 'inverse', 'freqs']):
            return modes

        matrix = None
        inverse = None
        basis = None
        freqs = None
        if isinstance(modes, dict):
            opts = modes.copy()
            for k in ["matrix", "modes"]:
                if k in opts:
                    matrix = opts[k]
                    del opts[k]
        elif hasattr(modes, 'matrix'):
            matrix = modes.matrix
            opts = {}

            for k in ['inverse', 'basis', 'freqs']:
                if hasattr(modes, k): opts[k] = getattr(modes, k)
        else:
            matrix = np.asanyarray(modes)
            opts = {}

        if 'inverse' in opts:
            inverse = opts['inverse']
            del opts['inverse']
        if 'basis' in opts:
            basis = opts['basis']
            del opts['basis']
        if 'freqs' in opts:
            freqs = opts['freqs']
            del opts['freqs']

        if matrix is None and inverse is None:
            raise ValueError(f"can't prep {cls.__name__} without matrix or inverse")

        if basis is None:
            if matrix.shape[0] % 3 == 0:
                basis = CartesianCoordinateSystem3D
            else:
                basis = InternalCoordinateSystem(dimension=(None, matrix.shape[0]))

        cls(
            basis,
            matrix,
            inverse=inverse,
            freqs=freqs,
            **opts
        )

    def __getitem__(self, item):
        """
        Takes a slice of the modes
        :param item:
        :type item:
        :return:
        :rtype:
        """

        if isinstance(item, (int, np.integer)):
            item = (item,)
        elif not isinstance(item[0], (int, np.integer)):
            item = tuple(item[0])

        sub_modes = self.matrix[:, item]
        inv = self._inv
        if inv is not None:
            inv = inv[item, :]
        freq = self.freqs[item,]
        return self.modify(
            matrix=sub_modes,
            freqs=freq,
            inverse=inv
        )

    def modify(self,
               matrix=None,
               *,
               freqs=None,
               origin=None,
               masses=None,
               inverse=None,
               name=None,
               mass_weighted=None,
               frequency_scaled=None,
               g_matrix=None
               ):
        return type(self)(
            self.basis,
            self.matrix if matrix is None else matrix,
            freqs=self.freqs if freqs is None else freqs,
            origin=self.origin if origin is None else origin,
            masses=self.masses if masses is None else masses,
            inverse=self.inverse if inverse is None else inverse,
            name=self.name if name is None else name,
            mass_weighted=self.mass_weighted if mass_weighted is None else mass_weighted,
            frequency_scaled=self.frequency_scaled if frequency_scaled is None else frequency_scaled,
            g_matrix=self.g_matrix if g_matrix is None else g_matrix,
        )

    def rotate(self, rot, in_place=False):
        raise NotImplementedError("too confusing...")

    def transform(self, tf, inv=None, origin=None):
        raise NotImplementedError("ambiguous...")
        #TODO: handle Cartesian variant where tf gets broadcasted

        if inv is None:
            if tf.shape[0] != tf.shape[1]:
                raise ValueError("if given only a transformation, need corresponding inverse")
            else:
                inv = np.linalg.inv(tf)
        base_inv = self.inverse
        base_mat = self.matrix
        raise Exception(base_inv @ base_mat, base_mat.shape, base_inv.shape)
        new_mat = tf@base_mat
        new_inv = base_inv@inv

        if origin is None:
            origin = self.origin
            if origin is not None:
                origin = tf@origin.flatten()

        return type(self)(
            self.basis,
            new_mat,
            freqs=self.freqs,
            masses=self.masses,
            origin=origin,
            inverse=new_inv
        )

    @property
    def cartesian_modes(self):
        return self.origin.ndim == 2

    def embed_coords(self, carts):
        flat_carts = (carts - self.origin[np.newaxis]).reshape((len(carts), -1))
        return (flat_carts[:, np.newaxis, :] @ self.inverse.T[np.newaxis]).reshape(
            flat_carts.shape[0],
            self.matrix.shape[1]
        )
    def unembed_coords(self, mode_coords):
        origin = self.origin
        carts = (mode_coords[:, np.newaxis, :] @ self.matrix.T[np.newaxis]).reshape(
            mode_coords.shape[:1] + origin.shape
        )
        carts = carts + origin[np.newaxis]
        return carts

    @property
    def total_transformation(self):
        return self._extended_coeffs
    @property
    def inverse_transformation(self):
        if self._inverse_coeffs is None:
            self._inverse_coeffs = nput.inverse_transformation(self.total_transformation,
                                                               len(self.total_transformation),
                                                               reverse_expansion=[self.inverse])
        return self._inverse_coeffs
    def embed_derivs(self, derivs):
        return nput.tensor_reexpand(self.total_transformation, derivs)
    def unembed_derivs(self, derivs):
        return nput.tensor_reexpand(self.inverse_transformation, derivs)

    @property
    def is_cartesian(self):
        if self.masses is not None:
            return self.matrix.shape[0] // 3 == len(self.masses)
        else:
            return 'Cartesian' in self.basis.name
    @property
    def coords_by_modes(self):
        return self.inverse
    @property
    def modes_by_coords(self):
        return self.matrix

    def _eval_G(self, masses):
        if not self.is_cartesian:  # a hack
            if self.origin is None:
                raise ValueError("can't get mass weighting matrix (G^-1/2) without structure")
            G = self.inverse.T @ self.inverse
            # raise NotImplementedError("will add at some point")
        else:
            G = np.diag(1 / np.repeat(masses, 3))
        return G

    def _get_gmatrix(self, masses=None):
        if masses is None:
            G = self.g_matrix
        else:
            G = None
        if G is None:
            if masses is None:
                masses = self.masses
                G = self.g_matrix = self._eval_G(self.masses)
            else:
                G = self._eval_G(self.masses)

        g12 = nput.fractional_power(G, 1/2)
        gi12 = nput.fractional_power(G, -1/2)
        return masses, g12, gi12

    def make_mass_weighted(self, masses=None):
        if self.mass_weighted: return self
        masses, g12, gi12 = self._get_gmatrix(masses=masses)
        L = g12 @ self.matrix
        Linv = self.inverse @ gi12
        origin = (self.origin.flatten()[np.newaxis, :] @ gi12).reshape(self.origin.shape)

        return self.modify(L,
                           inverse=Linv,
                           masses=masses,
                           origin=origin,
                           mass_weighted=True
                           )
    def remove_mass_weighting(self, masses=None):
        if not self.mass_weighted: return self
        masses, g12, gi12 = self._get_gmatrix(masses=masses)
        L = gi12 @ self.matrix
        Linv = self.inverse @ g12
        origin = (self.origin.flatten()[np.newaxis, :] @ g12).reshape(self.origin.shape)

        return self.modify(L,
                           inverse=Linv,
                           masses=masses,
                           origin=origin,
                           mass_weighted=False
                           )

    def _frequency_scaling(self, freqs=None):
        # L = self.matrix.shape.T
        if freqs is None:
            freqs = self.freqs
        conv = np.sqrt(freqs)
        return freqs, conv

    def make_frequency_scaled(self, freqs=None):
        if self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        L = self.matrix * conv[np.newaxis, :]
        Linv = self.inverse / conv[:, np.newaxis] # del_Q X

        return self.modify(L,
                           inverse=Linv,
                           freqs=freqs,
                           frequency_scaled=True
                           )

    def remove_frequency_scaling(self, freqs=None):
        if not self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        L = self.matrix / conv[np.newaxis, :]
        Linv = self.inverse * conv[:, np.newaxis] # del_Q X

        return self.modify(L,
                           inverse=Linv,
                           freqs=freqs,
                           frequency_scaled=False
                           )

    def make_dimensionless(self, freqs=None, masses=None):
        new = self
        if not new.mass_weighted: new = new.make_mass_weighted(masses=masses)
        if not new.frequency_scaled: new = new.make_frequency_scaled(freqs=freqs)

        return new

    def make_dimensioned(self, freqs=None, masses=None):
        new = self
        if not new.mass_weighted: new = new.remove_mass_weighting(masses=masses)
        if not new.frequency_scaled: new = new.remove_frequency_scaling(freqs=freqs)
        return new

    # def take_submodes(self, pos, axes=(0, 1, 2)):
    #     if self.is_cartesian:
    #         pos = np.asanyarray(pos)
    #         axes = np.asanyarray(axes)
    #         pos = (axes[np.newaxis, :] + pos[:, np.newaxis]*len(axes)).flatten()
    #     atom_disps = self.matrix[pos]
    #     f_base_sub = atom_disps.T @ np.diag(self.freqs**2) @ atom_disps
    #     return type(self).from_fg(
    #         self.basis.take_subbasis(pos),
    #         f_base_sub,
    #         self.g_matrix[np.ix_(pos, pos)],
    #         ...
    #     )



"""
Handles coordinate manipulation in a unified manner
"""
import numpy as np

from McUtils.Coordinerds import CoordinateSystem, CoordinateSystemConverter, CoordinateSystemConverters, ZMatrixCoordinates
import McUtils.Numputils as nput

class TotalCoordinateConverter:
    """
    A converter object that can apply normal mode transformations, non-linear transformations,
    and generic conversions (and does them in that order)
    """
    def __init__(self,
                 converter, inverse, reference_coords,
                 coordinate_expansion, inverse_expansion,
                 modes_matrix, modes_inverse
                 ):
        self.converter = converter
        self.inverse = inverse
        if converter is not None and inverse is None:
            raise ValueError("conversion function requires inverse")
        if inverse is not None and converter is None:
            raise ValueError("inverse conversion function requires forward function")
        self.ref = reference_coords
        self.coordinate_expansion, self.inverse_expansion = self.prep_expansions(coordinate_expansion, inverse_expansion)
        self.modes_matrix, self.modes_inverse = self.prep_modes(modes_matrix, modes_inverse)

    @classmethod
    def prep_modes(cls, modes_matrix, modes_inverse):
        if modes_matrix is not None and modes_inverse is None:
            modes_inverse = np.linalg.inv(modes_matrix)
        if modes_inverse is not None and modes_matrix is None:
            modes_matrix = np.linalg.inv(modes_inverse)

        return modes_matrix, modes_inverse

    @classmethod
    def prep_expansions(cls, coordinate_expansion, inverse_expansion):
        if coordinate_expansion is not None and inverse_expansion is None:
            inverse_expansion = nput.inverse_transformation(coordinate_expansion, len(coordinate_expansion))
        if inverse_expansion is not None and coordinate_expansion is None:
            coordinate_expansion = nput.inverse_transformation(inverse_expansion, len(inverse_expansion))

        return coordinate_expansion, inverse_expansion

    def convert(self, internals):
        internals = np.asanyarray(internals)
        smol = internals.ndim == 2
        if smol:
            internals = internals[np.newaxis]

        excess_shape = internals.shape[:-2]
        nc = internals.shape[-2]
        internals = internals.reshape(-1, internals.shape[-2]*internals.shape[-1])
        embedding_coords = [
            x for x in [0, 1, 2, 4, 5, 8]
            if x < internals.shape[-1]
        ]
        rem_coords = np.setdiff1d(np.arange(internals.shape[-1]), embedding_coords)
        embedding = internals[:, embedding_coords]
        internals = internals[:, rem_coords]
        if self.ref is not None:
            internals = internals - self.ref[np.newaxis]
        if self.modes_matrix is not None:
            internals = internals @ self.modes_matrix.T
        if self.coordinate_expansion is not None:
            coords = self.coordinate_expansion[0]
            for expansion_tensor in self.coordinate_expansion[1:]:
                contrib = expansion_tensor[np.newaxis]
                for _ in expansion_tensor.ndim:
                    contrib = nput.vec_tensordot(internals, expansion_tensor,
                                                 shared=1,
                                                 axes=[1, 1]
                                                 )
                coords += contrib
            internals = coords
        if self.converter is not None:
            internals = self.converter(internals)

        totals = np.empty((internals.shape[0], nc, 3), dtype=internals.dtype)
        totals[:, 0] = embedding[:, :3]
        if nc > 1:
            totals[:, 1, :1] = internals[:, 0]
            totals[:, 1, 1:] = embedding[:, 3:5]
        if nc > 2:
            totals[:, 2, :2] = internals[:, 1:3]
            totals[:, 2, 2] = embedding[:, 5]
        if nc > 3:
            totals[:, 3] = internals[:, 3:].reshape((internals.shape[0], nc-2, 3))

        if smol:
            totals = totals[0]

        return totals

    def invert(self, totals):
        totals = np.asanyarray(totals)
        smol = totals.ndim == 1
        if smol:
            totals = totals[np.newaxis]

        excess_shape = totals.shape[:-1]
        nc = totals.shape[-2]
        totals = totals.reshape(-1, totals.shape[-2] * totals.shape[-1])
        embedding_coords = [
            x for x in [0, 1, 2, 4, 5, 8]
            if x < totals.shape[-1]
        ]
        rem_coords = np.setdiff1d(np.arange(totals.shape[-1]), embedding_coords)
        embedding = totals[:, embedding_coords]
        totals = totals[:, rem_coords]

        if self.inverse is not None:
            totals = self.inverse(totals)

        if self.inverse_expansion is not None:
            coords = self.inverse_expansion[0]
            for expansion_tensor in self.inverse_expansion:
                contrib = expansion_tensor[np.newaxis]
                for _ in expansion_tensor.ndim:
                    contrib = nput.vec_tensordot(totals, expansion_tensor,
                                                 shared=1,
                                                 axes=[1, 1]
                                                 )
                coords += contrib
            totals = coords

        if self.modes_inverse is not None:
            totals = totals @ self.modes_inverse.T

        if self.ref is not None:
            totals = totals + self.ref[np.newaxis]

        internals = np.empty((totals.shape[0], nc, 3), dtype=totals.dtype)
        internals[:, 0] = embedding[:, :3]
        if nc > 1:
            internals[:, 1, :1] = totals[:, 0]
            internals[:, 1, 1:] = embedding[:, 3:5]
        if nc > 2:
            internals[:, 2, :2] = totals[:, 1:3]
            internals[:, 2, 2] = embedding[:, 5]
        if nc > 3:
            internals[:, 3] = totals[:, 3:].reshape((totals.shape[0], nc - 2, 3))

        internals = internals.reshape(excess_shape + (internals.shape[-1],))
        if smol:
            internals = internals[0]

        return internals

class TotalCoordinateSystem(CoordinateSystem):
    """
    A generic coordinate layer to unify coordinate handling,
    things written in terms of the basic `Coordinerds.CoordinateSet` should be
    converted over to this format bit-by-bit
    """

    # Design-wise, this is intended to solve the very common case
    # of having a set of coordinates that is defined as a function applied to a
    # set of base internal coordinates (or some linear combination of those base internal coordinates)
    # while still maintaining a reference (potentially) to some base set of Cartesian coordinates


    def __init__(self,
                 internals_spec,
                 *,
                 reference_coords=None,
                 converter=None,
                 inverse=None,
                 modes_matrix=None,
                 modes_inverse=None,
                 coordinate_expansion=None,
                 inverse_expansion=None,
                 **embedding_options
                 ):
        self.ints, self.embedding = self.canonicalize_internals(internals_spec, embedding_options)
        self.converter = TotalCoordinateConverter(
            converter, inverse, reference_coords,
            coordinate_expansion, inverse_expansion,
            modes_matrix, modes_inverse
        )
        super().__init__(
            "Total"+self.ints.name,
            dimension=(None, 3)
        )

    def canonicalize_internals(self, internals_spec, embedding):
        if isinstance(internals_spec, CoordinateSystem):
            embedding = dict(embedding, **internals_spec.converter_options)
        else:
            embedding['ordering'] = internals_spec
            internals_spec = ZMatrixCoordinates

        return internals_spec, embedding

class TotalCoordinatesToInternalCoordinateConverter(CoordinateSystemConverter):
    def __init__(self, totals:TotalCoordinateSystem, internals, **opts):
        self._types = (totals, internals)
        super().__init__(**opts)
    @property
    def types(self):
        return self._types
    def convert(self, coords, **kw):
        return self._types[0].converter.invert(coords)

    def convert_many(self, coords, **kwargs):
        return self._types[0].converter.invert(coords)

class InternalCoordinateToTotalCoordinatesConverter(CoordinateSystemConverter):
    def __init__(self, totals: TotalCoordinateSystem, internals, **opts):
        self._types = (internals, totals)
        super().__init__(**opts)

    @property
    def types(self):
        return self._types

    def convert(self, coords, **kw):
        return self._types[0].converter.convert(coords)

    def convert_many(self, coords, **kwargs):
        return self._types[0].converter.convert(coords)



"""
Provides utilities for converting state vectors to indices.
Used implicitly by various conversion utilities
"""

import abc, numpy as np, collections
from McUtils.Combinatorics import SymmetricGroupGenerator

__all__ = [
    "BaseStateIndexer",
    "ArrayStateIndexer",
    "SpaceStateIndexer",
    "PermutationStateIndexer"
]

class BaseStateIndexer(metaclass=abc.ABCMeta):
    """
    Provides a base class for indexing states
    """
    @abc.abstractmethod
    def to_indices(self, states):
        """
        Converts the set of states to numerical indices

        :param states:
        :type states:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract base class")
    @abc.abstractmethod
    def from_indices(self, indices):
        """
        Converts the set of states to numerical indices

        :param states:
        :type states:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract base class")

class ArrayStateIndexer(BaseStateIndexer):
    """
    Very simple indexer that takes a set of max state dimensions and
    provides the appropriate indices in that space
    """
    def __init__(self, dims):
        self.dims = np.array(dims)

    def to_state(self, serializer=None):
        return {
            'dims': self.dims
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(data['dims'])

    def to_indices(self, states):
        idx = np.asarray(states, dtype=int)
        if len(states) == 0:
            return idx
        if idx.ndim == 1:
            idx = idx[np.newaxis]
        return np.ravel_multi_index(idx.T, self.dims)
    def from_indices(self, indices):
        if len(indices) == 0:
            return np.array([], dtype='int8')
        unravel = np.unravel_index(indices, self.dims)
        return np.array(unravel, dtype='int8').T

class SpaceStateIndexer(BaseStateIndexer):
    """
    Very simple indexer that takes a set of states and indexes based on that.
    Needs some testing to make sure everything works as cleanly as hoped.
    """
    def __init__(self, states):
        self.og_states = np.asarray(states, dtype=int)
        self.indexer = np.lexsort(states.T)
        if self.og_states.ndim != 2:
            raise ValueError("can't index anything other than vectors of excitations")
    def to_indices(self, states):
        if len(states) == 0:
            return np.array([], dtype=int)
        return np.searchsorted(self.og_states, states, indexer=self.indexer)
    def from_indices(self, indices):
        return self.og_states[indices,]

PermutationStateKey = collections.namedtuple("PermutationStateKey", ['non_zero', 'classes'])
class PermutationStateIndexer(BaseStateIndexer):
    """
    A sophisticated indexer that takes a state dimension and provides
    indices based on the `shortlex` ordering, where ordering is defined
    first by # of quanta of excitation, then by which partitioning of the #quanta,
     it represents, and finally by which permutation of that paritioning it is.
    Is likely about as stable as an indexer can be expected to be over large
    numbers of states. Unlikely to exhaust the max integers available for most
    systems.
    """
    def __init__(self, ndim):
        self.ndim = ndim
        self.symmetric_group = SymmetricGroupGenerator(self.ndim)
        self.partitioning_cache = {}

    def to_state(self, serializer=None):
        return {
            'ndim': self.ndim
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(data['ndim'])

    def to_indices(self, states):
        """
        Finds the appropriate integer partitioning for each state
        :param states: 2D array of states as excitations
        :type states: np.ndarray
        :return:
        :rtype:
        """

        return self.symmetric_group.to_indices(states)

    def from_indices(self, indices, check=False):
        """
        Inverts the index calculation.
        First determines what number of quanta the index corresponds to,
        then which integer partition, and finally just loops through the unique
        permutations of the partition to get the right one.
        This is not assured to be a fast process in any way.

        :param indices:
        :type indices: Iterable[int]
        :return:
        :rtype:
        """

        return self.symmetric_group.from_indices(indices)


